# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import dataclasses
import datetime
import uuid
from string import Template
from typing import Annotated, Self

import dagger
import yaml
from dagger import BuildArg, Doc, Ignore, dag, function, object_type

from .hpc import Singularity
from .interlink import InterLink
from .k8s import create_pod_manifest
from .literals import MLFramework
from .utils import get_codename

ignore_list = [
    "**",
    "!src/itwinai/**",
    "!tests/**",
    "!env-files/**",
    "!use-cases/**",
    "**/__pycache__",
    "!pyproject.toml",
]


@object_type
class Itwinai:
    """CI/CD for container images of itwinai."""

    docker_registry: str
    singularity_registry: str
    image: str
    tag: str | None
    container: dagger.Container | None = dagger.field(default=None)
    nickname: str

    _unique_id: str | None = dataclasses.field(default=None, init=False)
    _logs: list[str] = dataclasses.field(default_factory=list, init=False)

    @classmethod
    def create(
        cls,
        docker_registry: Annotated[
            str,
            Doc("The Docker registry base URL where the container will be published"),
        ] = "ghcr.io/intertwin-eu",
        singularity_registry: Annotated[
            str,
            Doc(
                "Harbor registry namespace (i.e., 'registry/project') where to publish the "
                "Singularity images"
            ),
        ] = "registry.egi.eu/dev.intertwin.eu",
        image: Annotated[
            str,
            Doc("The name of the container image"),
        ] = "itwinai-dev",
        nickname: Annotated[
            str,
            Doc(
                "A simple name to indicate the flavor of the image. Used to generate the "
                "corresponding 'latest' tag"
            ),
        ] = "torch",
        tag: Annotated[
            str | None,
            Doc("Tag for the container image. Defaults to random uuid if not provided"),
        ] = None,
        container: Annotated[
            dagger.Container | None,
            Doc("Optional container image to use as itwinai container"),
        ] = None,
    ):
        tag = tag or str(uuid.uuid4())
        return cls(
            singularity_registry=singularity_registry,
            docker_registry=docker_registry,
            image=image,
            tag=tag,
            container=container,
            nickname=nickname,
        )

    @function
    def logs(self) -> str:
        """Print the logs generted so far. Useful to terminate a chain of steps
        returning Self, preventing lazy execution of it.
        """
        return "\n\n".join(self._logs) or "There are no logs to show"

    @function
    async def build_container(
        self,
        context: Annotated[
            dagger.Directory,
            Ignore(ignore_list),
            Doc("location of source directory"),
        ],
        dockerfile: Annotated[
            str,
            Doc("location of Dockerfile"),
        ],
        build_args: Annotated[
            list[str] | None,
            Doc("Comma-separated build args"),
        ] = None,
    ) -> Self:
        """Build itwinai container image from existing Dockerfile"""
        if build_args:
            build_args = [
                BuildArg(name=arg_couple.split("=")[0], value=arg_couple.split("=")[1])
                for arg_couple in build_args
            ]

        self.container = (
            dag.container()
            .build(
                context=context,
                dockerfile=dockerfile,
                build_args=build_args,
            )
            .with_label(
                name="org.opencontainers.image.created",
                value=datetime.datetime.now(datetime.timezone.utc).isoformat(),
            )
        )

        # Get itwinai version
        itwinai_version = (
            await self.container.with_exec(
                [
                    "bash",
                    "-c",
                    'pip list | grep -w "itwinai" | head -n 1 | tr -s " " | cut -d " " -f2 ',
                ]
            ).stdout()
        ).strip()
        self._logs.append(f"INFO: Built container for itwinai v{itwinai_version}")

        self.container = self.container.with_label(
            name="org.opencontainers.image.version",
            value=itwinai_version,
        )

        return self

    @function
    async def test_local(
        self, cmd: Annotated[list[str] | None, Doc("Command to run tests")] = None
    ) -> Self:
        """Test itwinai container image with pytest on non-HPC environments."""

        cmd = cmd or [
            "pytest",
            "-v",
            "--disable-warnings",
            # tell pytest-xdist to parallelize tests over logical cores
            "-n",
            "logical",
            "-m",
            "not hpc and not functional and not tensorflow",
            "/app/tests",
        ]

        tests_result = await self.container.with_exec(cmd).stdout()
        self._logs.append("INFO: running pytest for 'local' tests:")
        self._logs.append(tests_result)
        return self

    @function
    async def publish(
        self,
        uri: Annotated[
            str | None,
            Doc("Optional target URI for the image"),
        ] = None,
    ) -> Self:
        """Push container to registry"""

        uri = uri or f"{self.docker_registry}/{self.image}:{self.tag}"
        outcome = await (
            self.container.with_label(
                name="org.opencontainers.image.ref.name",
                value=uri,
            )
            # Invalidate cache to ensure that the container is always pushed
            .with_env_variable(
                "CACHE", datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")
            )
            .publish(address=uri)
        )
        self._logs.append(f"INFO: publishing Docker image to {uri}")
        self._logs.append(outcome)
        return self

    @function
    def interlink(
        self,
        values: Annotated[dagger.Secret, Doc("interLink installer configuration")],
        name: Annotated[
            str, Doc("Name of the k3s cluster in which the interLink VK is deployed")
        ] = "interlink-cluster",
        wait: Annotated[
            int,
            Doc(
                "Sleep time (seconds) needed to wait for the VK to appear in the k3s cluster."
            ),
        ] = 60,
        vk_name: Annotated[
            str | None,
            Doc("Name of the interLink VK. Automatically detected from values if not given"),
        ] = None,
        kubernetes: Annotated[
            dagger.Service | None,
            Doc(
                "Endpoint to exsisting k3s service to attach to. "
                "Example (from the CLI): tcp://localhost:1997"
            ),
        ] = None,
    ) -> InterLink:
        """Get interLink service."""
        return InterLink(
            values=values, name=name, wait=wait, vk_name=vk_name, kubernetes=kubernetes
        )

    @function
    async def test_hpc(
        self,
        values: Annotated[dagger.Secret, Doc("interLink installer configuration")],
        name: Annotated[
            str, Doc("Name of the k3s cluster in which the interLink VK is deployed")
        ] = "interlink-cluster",
        wait: Annotated[
            int,
            Doc(
                "Sleep time (seconds) needed to wait for the VK to appear in the k3s cluster."
            ),
        ] = 60,
        kubernetes: Annotated[
            dagger.Service | None,
            Doc(
                "Endpoint to exsisting k3s service to attach to. "
                "Example (from the CLI): tcp://localhost:1997"
            ),
        ] = None,
        image: Annotated[
            str | None,
            Doc(
                "Container image to test on HPC. If given, it will override any itwinai "
                "container previously created by other functions in the chain. Example: "
                "docker://ghcr.io/intertwin-eu/itwinai:latest"
            ),
        ] = None,
    ) -> Self:
        """Test container on remote HPC using interLink. Note that the container image must
        already exist in some publicly accessible containers registry.
        """

        image = image or f"docker://{self.docker_registry}/{self.image}:{self.tag}"
        self._logs.append(f"INFO: testing on HPC image {image}")

        # Create pod manifest
        gpus_per_node = 1
        cpus_per_gpu = 4
        num_nodes = 2
        jobscript = "/app/tests/torch/slurm.vega.sh"
        pre_exec_cmd = (
            "export CONTAINER_PATH=itwinai_dist_test.sif "
            f"&& singularity pull --force $CONTAINER_PATH {image} "
            f"&& singularity exec $CONTAINER_PATH cat {jobscript} > slurm.vega.sh "
            # Activate env on Vega
            "&& . /etc/bashrc "
            "&& source /ceph/hpc/software/cvmfs_env.sh "
            "&& module use /usr/share/Modules/modulefiles "
            # Quick fix
            # f"&& export SLURM_GPUS_PER_NODE={gpus_per_node} "
            # f"&& export SLURM_CPUS_PER_GPU={cpus_per_gpu} "
            # Env variables necessary for tests
            "&& export MNIST_PATH=/ceph/hpc/data/st2301-itwin-users/mbunino/mnist "
            "&& export NO_COLOR=1 "
            "&& export SHARED_FS_PATH=/ceph/hpc/data/st2301-itwin-users/tmp-mbunino2 "
            # Launch code in SLURM job
            # Ray
            "&& export DIST_MODE=ray "
            "&& export RUN_NAME=ray-itwinai "
            "&& export COMMAND='pytest -v -m ray_dist /app/tests' "
            "&& source slurm.vega.sh "
            # DDP
            "&& export DIST_MODE=ddp "
            "&& export RUN_NAME=ddp-itwinai "
            "&& export COMMAND='pytest -v -m torch_dist /app/tests' "
            "&& source slurm.vega.sh "
            # DeepSpeed
            "&& export DIST_MODE=deepspeed "
            "&& export RUN_NAME=ds-itwinai "
            "&& export COMMAND='pytest -v -m deepspeed_dist /app/tests' "
            "&& source slurm.vega.sh "
            # Horovod
            "&& export DIST_MODE=horovod "
            "&& export RUN_NAME=horovod-itwinai "
            "&& export COMMAND='pytest -v -m horovod_dist /app/tests' "
            "&& source slurm.vega.sh "
            # Ray: must be at the bottom! srun are blocking forever and won't release
            # the resources, this no more srun cannot be invoked after a Ray cluster
            # is launched.
            "&& export DIST_MODE=ray "
            "&& export RUN_NAME=ray-itwinai "
            "&& export COMMAND='pytest -v -m ray_dist /app/tests' "
            "&& source slurm.vega.sh "
        )
        # Request memory and CPUs
        resources = {
            "limits": {"cpu": 48, "memory": "150Gi"},
            "requests": {"cpu": cpus_per_gpu * gpus_per_node, "memory": "20Gi"},
        }
        hpc_partition = "gpu"
        annotations = {
            "slurm-job.vk.io/flags": (
                # --cpus-per-gpu fails on Vega through interLink
                f"-p {hpc_partition} --gres=gpu:{gpus_per_node} "
                f"--gpus-per-node={gpus_per_node} "
                f"--ntasks-per-node=1 --nodes={num_nodes} "
                f"--cpus-per-task={cpus_per_gpu * gpus_per_node} "
                "--time=01:00:00 "
            ),
            "slurm-job.vk.io/pre-exec": (
                "trap 'export SINGULARITYENV_PRE_EXEC_RETURN_CODE=1' ERR && "
                f"(set -e && {pre_exec_cmd}) ; trap - ERR"
            ),
        }
        image_path = "/ceph/hpc/data/st2301-itwin-users/cern/hello-world-image.sif"
        cmd_args = ["sleep 10 && exit $PRE_EXEC_RETURN_CODE"]
        pod_name = f"ci-test-itwinai-hpc-{self.image}-{self.tag}"

        # Launch interLink service
        interlink_svc = self.interlink(
            values=values, name=name, wait=wait, kubernetes=kubernetes
        )
        async with interlink_svc.start_serving():
            pod_manifest = create_pod_manifest(
                annotations=annotations,
                image_path=image_path,
                cmd_args=cmd_args,
                name=pod_name,
                resources=resources,
                target_node=interlink_svc.vk_name,
            )
            pod_manifest_str = yaml.dump(pod_manifest)
            self._logs.append(f"INFO: submitting the following pod:\n{pod_manifest_str}")

            # self._logs.append("DEBUG: interlink cluster status")
            # self._logs.append(await interlink_svc.status())

            self._logs.append(await interlink_svc.submit_pod(pod_manifest_str))
            status, logs = await interlink_svc.wait_pod(
                name=pod_name, timeout=10000, poll_interval=600
            )

        if status not in ["Succeeded", "Completed"]:
            message = (
                f"Pod did not complete successfully! Status: {status}\n"
                f"{'#' * 100}\nJOB LOGS:\n\n{logs}"
            )
            print(message)
            raise RuntimeError(message)

        self._logs.append(f"INFO: pod finished with status: '{status}'")
        self._logs.append(f"INFO: pod finished with logs: \n{logs}")

        return self

    @function
    async def dev_pipeline(
        self,
        tag_template: Annotated[
            str,
            Doc("Custom image tag pattern."),
        ] = "${itwinai_version}-torch${framework_version}-${os_version}",
        framework: Annotated[
            MLFramework, Doc("ML framework in container")
        ] = MLFramework.TORCH,
        skip_singularity: Annotated[bool, Doc("Avoid publishing a Singularity image")] = False,
        password: Annotated[
            dagger.Secret | None, Doc("Password for Singularity registry")
        ] = None,
        username: Annotated[
            dagger.Secret | None, Doc("Username for Singularity registry")
        ] = None,
    ) -> str:
        """CI pipeline for pre-release containers. Tests are only local."""

        if not skip_singularity:
            assert username is not None, "Missing username for Singularity registry"
            assert password is not None, "Missing password for Singularity registry"

        # Test locally
        await self.test_local()

        # Publish to Docker registry with "final" tag
        tag = await self._evaluate_tag_template(tag_template=tag_template, framework=framework)
        await self.publish(uri=f"{self.docker_registry}/{self.image}:{tag}")

        if not skip_singularity:
            # Publish to Singularity registry
            await self.publish_singularity(
                uri=f"oras://{self.singularity_registry}/{self.image}:{tag}",
                username=username,
                password=password,
            )

        return self.logs()

    @function
    async def release_pipeline(
        self,
        values: Annotated[dagger.Secret, Doc("interLink installer configuration")],
        tag_template: Annotated[
            str | None,
            Doc(
                "Custom image tag pattern. Example: "
                "'${itwinai_version}-torch${framework_version}-${os_version}'"
            ),
        ] = None,
        framework: Annotated[
            MLFramework, Doc("ML framework in container")
        ] = MLFramework.TORCH,
        skip_hpc: Annotated[bool, Doc("Skip tests on remote HPC")] = False,
        skip_singularity: Annotated[bool, Doc("Avoid publishing a Singularity image")] = False,
        kubernetes: Annotated[
            dagger.Service | None,
            Doc(
                "Endpoint to exsisting k3s service to attach to. "
                "Example (from the CLI): tcp://localhost:1997"
            ),
        ] = None,
        interlink_cluster_name: Annotated[
            str, Doc("Name of the k3s cluster in which the interLink VK is deployed")
        ] = "interlink-cluster",
        password: Annotated[
            dagger.Secret | None, Doc("Password for Singularity registry")
        ] = None,
        username: Annotated[
            dagger.Secret | None, Doc("Username for Singularity registry")
        ] = None,
    ) -> str:
        """CI pipeline for release containers. Test on HPC and generate Singularity images."""

        if not skip_singularity:
            assert username is not None, "Missing username for Singularity registry"
            assert password is not None, "Missing password for Singularity registry"

        # Test locally
        await self.test_local()

        if not skip_hpc:
            if skip_singularity:
                # The Docker image will be automatically converted to Singularity on HPC
                # before running the tests
                uri = f"{self.docker_registry}/itwinai-dev:{self.tag}"
                await self.publish(uri=uri)
                # Append Docker prefix to tell singularity where to pull the image from
                uri = f"docker://{uri}"
            else:
                # Publish to registry with random hash
                uri = f"oras://{self.singularity_registry}/itwinai-dev:{self.tag}"
                await self.publish_singularity(uri=uri, username=username, password=password)

            # Test on HPC with
            await self.test_hpc(
                values=values, kubernetes=kubernetes, name=interlink_cluster_name, image=uri
            )

        else:
            self._logs.append("INFO: skipping tests on HPC")

        # Publish to registry with "final" and latest tag
        final_tag = await self._evaluate_tag_template(
            tag_template=tag_template, framework=framework
        )
        for tag in [final_tag, f"{self.nickname}-latest"]:
            # Publish to Docker registry
            await self.publish(uri=f"{self.docker_registry}/{self.image}:{tag}")
            if not skip_singularity:
                # Publish to Singularity registry
                await self.publish_singularity(
                    uri=f"oras://{self.singularity_registry}/{self.image}:{tag}",
                    username=username,
                    password=password,
                )

        return self.logs()

    async def _evaluate_tag_template(self, tag_template: str, framework: MLFramework) -> str:
        """Interpolate the fields in the tag template with values computed from the
        container.
        """
        itwinai_version = (
            await self.container.with_exec(
                [
                    "bash",
                    "-c",
                    'pip list | grep -w "itwinai" | head -n 1 | tr -s " " | cut -d " " -f2 ',
                ]
            ).stdout()
        ).strip()
        os_info = (
            await self.container.with_exec(
                [
                    "bash",
                    "-c",
                    "cat /etc/*-release",
                ]
            ).stdout()
        ).strip()
        os_version = get_codename(os_info)

        if framework == MLFramework.TORCH:
            tag_template = (
                tag_template or "${itwinai_version}-torch${framework_version}-${os_version}"
            )
            framework_version = (
                await self.container.with_exec(
                    [
                        "bash",
                        "-c",
                        (
                            'pip list | grep -w "torch[^-]" | head -n 1 | tr -s " " '
                            '| cut -d " " -f2 | cut -f1,2 -d .'
                        ),
                    ]
                ).stdout()
            ).strip()
        elif framework == MLFramework.TENSORFLOW:
            tag_template = (
                tag_template or "${itwinai_version}-tf${framework_version}-${os_version}"
            )
            framework_version = (
                await self.container.with_exec(
                    [
                        "bash",
                        "-c",
                        (
                            # TODO: check this command
                            'pip list | grep -w "tensorflow[^-]" | head -n 1 | tr -s " " '
                            '| cut -d " " -f2 | cut -f1,2 -d .'
                        ),
                    ]
                ).stdout()
            ).strip()

        # In GH actions ${...} is interpolated, so we replace $ with @
        return Template(tag_template.replace("@", "$")).substitute(
            itwinai_version=itwinai_version,
            framework_version=framework_version,
            os_version=os_version,
        )

    @function
    def singularity(
        self,
        base_image: Annotated[
            dagger.Container, Doc("Base Singularity image")
        ] = dag.container().from_("quay.io/singularity/docker2singularity"),
        docker: Annotated[
            dagger.Container | None,
            Doc(
                "Docker container to convert to Singularity. "
                "If given it overrides the current itwinai container."
            ),
        ] = None,
    ) -> Singularity:
        """Access Singularity module."""
        return Singularity(base_image=base_image, docker=docker or self.container)

    @function
    async def publish_singularity(
        self,
        password: Annotated[dagger.Secret, Doc("Password for Singularity registry")],
        username: Annotated[dagger.Secret, Doc("Username for Singularity registry")],
        uri: Annotated[
            str | None,
            Doc("Optional target URI for the image"),
        ] = None,
    ) -> Self:
        """Convert itwinai container to Singularity and push it to some registry."""
        uri = uri or f"oras://{self.singularity_registry}/{self.image}:{self.tag}"
        self._logs.append(f"INFO: the Singularity image will be published to: {uri}")
        stdout = await self.singularity(docker=self.container).publish(
            password=password, username=username, uri=uri
        )
        self._logs.append(stdout)
        return self



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------


def get_codename(release_info: str) -> str:
    """
    Extracts the codename (VERSION_CODENAME or os_version) from release information.

    Args:
        release_info (str): The string containing the output of /etc/*-release.

    Returns:
        str: The extracted codename (e.g., "jammy" or "bookworm").
    """
    # Create a dictionary from the release info
    release_dict = {}
    for line in release_info.splitlines():
        if "=" in line:
            key, value = line.split("=", 1)
            release_dict[key.strip()] = value.strip().strip('"')

    # Attempt to extract the codename
    return release_dict.get("VERSION_CODENAME", release_dict.get("os_version", "Unknown"))



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# - Anna Lappe <anna.elisa.lappe@cern.ch> - CERN
# -------------------------------------------------------------------------------------

import tempfile
from pathlib import Path
from unittest.mock import MagicMock

import pytest

from itwinai.loggers import (
    ConsoleLogger,
    Logger,
    LoggersCollection,
    MLFlowLogger,
    Prov4MLLogger,
    TensorBoardLogger,
    WandBLogger,
)
from itwinai.torch.loggers import ItwinaiLogger as PyTorchLightningLogger


@pytest.fixture
def console_logger():
    with tempfile.TemporaryDirectory() as temp_dir:
        save_dir = Path(temp_dir) / "console/test_mllogs"
        yield ConsoleLogger(savedir=save_dir, log_freq=1)


@pytest.fixture(scope="module")
def mlflow_logger():
    with tempfile.TemporaryDirectory() as temp_dir:
        save_dir = Path(temp_dir) / "mlflow/test_mllogs"
        yield MLFlowLogger(
            savedir=save_dir,
            experiment_name="test_experiment",
        )


@pytest.fixture
def wandb_logger():
    with tempfile.TemporaryDirectory() as temp_dir:
        save_dir = Path(temp_dir) / "wandb/test_mllogs"
        yield WandBLogger(savedir=save_dir, project_name="test_project", offline_mode=True)


@pytest.fixture
def tensorboard_logger_tf():
    with tempfile.TemporaryDirectory() as temp_dir:
        save_dir = Path(temp_dir) / "tf_tb/test_mllogs"
        yield TensorBoardLogger(savedir=save_dir, framework="tensorflow")


@pytest.fixture
def tensorboard_logger_torch():
    with tempfile.TemporaryDirectory() as temp_dir:
        save_dir = Path(temp_dir) / "torch_tb/test_mllogs"
        yield TensorBoardLogger(savedir=save_dir, framework="pytorch")


@pytest.fixture
def prov4ml_logger():
    return Prov4MLLogger()


@pytest.fixture
def loggers_collection(
    console_logger,
    mlflow_logger,
    wandb_logger,
    tensorboard_logger_torch,
    prov4ml_logger,
):
    return LoggersCollection(
        [
            console_logger,
            mlflow_logger,
            wandb_logger,
            tensorboard_logger_torch,
            prov4ml_logger,
        ]
    )


@pytest.fixture
def lightning_mock_loggers():
    """Setup a generic PyTorchLightningLogger with mock loggers for testing."""
    itwinai_logger_mock = MagicMock(spec=Logger)
    # Setting default attributes as per class definition
    itwinai_logger_mock.is_initialized = False
    itwinai_logger_mock.worker_rank = 0
    lightning_logger = PyTorchLightningLogger(itwinai_logger=itwinai_logger_mock)

    return itwinai_logger_mock, lightning_logger



import sys
from normflow import Model, Fitter
from normflow.nn import DistConvertor_
from normflow.action import ScalarPhi4Action
from normflow.prior import NormalPrior

def make_model():
    net_ = DistConvertor_(10, symmetric=True)
    prior = NormalPrior(shape=(1,))
    action = ScalarPhi4Action(kappa=0, m_sq=-1.2, lambd=0.5)

    return Model(net_=net_, prior=prior, action=action)

def fit_func(model, n_epochs=100, strategy='ddp'):
    """Training function to fit model."""

    config = {
        "optim_lr": 0.001,
        "weight_decay": 0.01,
        "ckpt_disp": False,
        "batch_size": 128,
        "save_every": "None",
        "optimizer_class": "torch.optim.AdamW",
        "scheduler": "None",
        "loss_fn": "None",
        "print_stride": 10,
        "print_batch_size": 1024,
        "snapshot_path": None,
        "epochs_run": 0
    }
    # Initialize the Fitter and execute the training
    fitter = Fitter(model=model, epochs=n_epochs, config=config, strategy=strategy)
    fitter.execute()

def main():
    model = make_model()
    fit_func(model)

if __name__ == "__main__":
    main()



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Alex Krochak
#
# Credit:
# - Alex Krochak <o.krochak@fz-juelich.de> - FZJ
# --------------------------------------------------------------------------------------

import os
from typing import Any, Dict, Literal, Optional, Tuple
from torch.utils.data import Dataset

import torch.nn as nn
from torch import save
from itwinai.torch.trainer import TorchTrainer
from itwinai.torch.config import TrainingConfiguration
from itwinai.loggers import Logger
from itwinai.torch.monitoring.monitoring import measure_gpu_utilization
from itwinai.torch.profiling.profiler import profile_torch_trainer

class PulsarTrainer(TorchTrainer):
    """Trainer class for radio-astronomy use-case.
    Inherited from itwinai TorchTrainer class."""

    def __init__(
        self,
        config: Dict | TrainingConfiguration | None = None,
        strategy: Literal["ddp", "deepspeed", "horovod"] | None = None,
        logger: Logger | None = None,
        epochs: int = 3,
        model: nn.Module = None,
        loss: nn.Module = None,
        store_trained_model_at: str = ".models/model.pt",
        name: Optional[str] = None,
        measure_gpu_data: bool = False,
        measure_communication_overhead: bool = False,
        measure_epoch_time: bool = False,
    ) -> None:

        # these parameters are initialized with the TorchTrainer class:
        super().__init__(
            config=config,
            strategy=strategy,
            logger=logger,
            epochs=epochs,
            model=model,
            name=name,
            measure_gpu_data=measure_gpu_data,
            measure_communication_overhead=measure_communication_overhead,
            measure_epoch_time=measure_epoch_time,
        )
        # set the custom loss function
        self.loss = loss
        self.store_trained_model_at = store_trained_model_at
        os.makedirs(os.path.dirname(store_trained_model_at), exist_ok=True)

    # NOTE: this would be a nice way to re-use the original __init__
    # and insert the custom loss function, but AFAIK this doesn't
    # work nicely while running from config.yaml

    # def set_attributes(self, loss, store_trained_model_at) -> None:
    #     self.loss = loss
    #     self.store_trained_model_at = store_trained_model_at
    #     os.makedirs(os.path.dirname(store_trained_model_at), exist_ok=True)

    def create_model_loss_optimizer(self) -> None:

        ### This code is almost a complete copy of this method from ###
        ### src/itwinai/torch/trainer.py, with exception of removed ###
        ### loss function definition as it is already set in the    ###
        ### constructor                                             ###

        # Model, optimizer, and lr scheduler may have already been loaded from a checkpoint

        if self.model is None:
            raise ValueError(
                "self.model is None! Either pass it to the constructor, load a checkpoint, or "
                "override create_model_loss_optimizer method."
            )
        if self.model_state_dict:
            # Load model from checkpoint
            self.model.load_state_dict(self.model_state_dict, strict=False)

        # Parse optimizer from training configuration
        # Optimizer can be changed with a custom one here!
        self._set_optimizer_from_config()

        # Parse LR scheduler from training configuration
        # LR scheduler can be changed with a custom one here!
        self._set_lr_scheduler_from_config()

        if self.optimizer_state_dict:
            # Load optimizer state from checkpoint
            # IMPORTANT: this must be after the learning rate scheduler was already initialized
            # by passing to it the optimizer. Otherwise the optimizer state just loaded will
            # be modified by the lr scheduler.
            self.optimizer.load_state_dict(self.optimizer_state_dict)

        if self.lr_scheduler_state_dict and self.lr_scheduler:
            # Load LR scheduler state from checkpoint
            self.lr_scheduler.load_state_dict(self.lr_scheduler_state_dict)

        # IMPORTANT: model, optimizer, and scheduler need to be distributed from here on

        distribute_kwargs = self.get_default_distributed_kwargs()

        # Distributed model, optimizer, and scheduler
        (self.model, self.optimizer, self.lr_scheduler) = self.strategy.distributed(
            self.model, self.optimizer, self.lr_scheduler, **distribute_kwargs
        )

    def write_model(self) -> None:
        """Write the model to disk."""
        save(self.model.state_dict(), self.store_trained_model_at)

    # Hacky way to get around the fact that the execute method
    # does not return self.model anymore
    @profile_torch_trainer
    @measure_gpu_utilization
    def execute(
        self,
        train_dataset: Dataset,
        validation_dataset: Dataset | None = None,
        test_dataset: Dataset | None = None,
    ) -> Tuple[Dataset, Dataset, Dataset, Any]:
        objs = super().execute(
            train_dataset=train_dataset,
            validation_dataset=validation_dataset,
            test_dataset=test_dataset,
        )
        # return the datasets and the model
        return *(objs[:-1]), self.model



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Javad Komijani - ETHZ
#
# Credit:
# - Javad Komijani <jkomijani@gmail.com> - ETHZ
# - Gaurav Sinha Ray <sinha@ifca.unican.es> - CSIC
# - Rakesh Sarma <r.sarma@fz-juelich.de> - Juelich
# --------------------------------------------------------------------------------------

# Copyright (c) 2021-2024 Javad Komijani

"""This module contains high-level classes for normalizing flow techniques,
with the central `Model` class integrating essential components such as priors,
networks, and actions. It provides utilities for training and sampling,
along with support for MCMC sampling. It also integrates with the itwinai package
to perform distributed training, profiling and logging.
"""

import torch
import torch.distributed as dist
import time
import logging, os, sys
from pathlib import Path
from typing import Any, Callable, Dict, Literal, Optional, Tuple
from torch.utils.data import Dataset
import torch.optim.lr_scheduler as lr_scheduler
import numpy as np

from .mcmc import MCMCSampler, BlockedMCMCSampler
from .lib.combo import estimate_logz, fmt_val_err
from normflow.prior import Prior
from normflow.nn import ModuleList_
from normflow.action.scalar_action import ScalarPhi4Action

from itwinai.torch.trainer import TorchTrainer
from itwinai.torch.distributed import DeepSpeedStrategy
from itwinai.loggers import EpochTimeTracker, Logger
from itwinai.torch.monitoring.monitoring import measure_gpu_utilization
from itwinai.torch.profiling.profiler import profile_torch_trainer
from itwinai.distributed import suppress_workers_print
from itwinai.torch.config import TrainingConfiguration


class Model:
    """The central high-level class of the package, which integrates instances of
    essential classes (`prior`, `net_`, and `action`) to provide utilities for
    training and sampling. This class interfaces with various core components
    to facilitate training, posterior inference and MCMC sampling.

    Args:
        prior (Prior): instance of a `Prior` class
            An instance of a Prior class (e.g., `NormalPrior`) representing the
            model's prior distribution.
        net_ (Module_): instance of a `Module_` class
            A model component responsible for the transformations required in the
            model. The trailing underscore indicates that the associated forward
            method computes and returns the Jacobian of the transformation, which
            is crucial in the method of normalizing flows.
        action (Action): instance of an `ScalarPhi4Action` class
            Defines the model's action, which specified the target distribution
            during training.

    Attributes:
        posterior (Posterior): An instance of the Posterior class,
            which manages posterior inference on the model parameters.
        mcmc (MCMCSampler): An instance of the MCMCSampler class, enabling
            MCMC sampling for posterior distributions.
        blocked_mcmc (BlockedMCMCSampler): An instance of the BlockedMCMCSampler class,
            providing blockwise MCMC sampling for improved sampling efficiency.
    """

    def __init__(self, *, prior: Prior, net_: ModuleList_, action: ScalarPhi4Action):
        self.net_ = net_
        self.prior = prior
        self.action = action

        self.posterior = Posterior(self)
        self.mcmc = MCMCSampler(self)
        self.blocked_mcmc = BlockedMCMCSampler(self)


class Posterior:
    """Creates samples directly from a trained probabilistic model.
    The `Posterior` class generates samples from a specified model without
    using an accept-reject step, making it suitable for tasks that require
    quick, direct sampling. All methods in this class use `torch.no_grad()`
    to prevent gradient computation.

    Args:
        model (Model): A trained model to sample from.
    """

    def __init__(self, model: Model):
        self._model = model

    @torch.no_grad()
    def sample(self, batch_size: int = 1, **kwargs) -> torch.Tensor:
        """Draws samples from the model.

        Args:
            batch_size (int, optional): Number of samples to draw. Defaults to 1.

        Returns:
            torch.Tensor: Generated samples.
        """
        return self.sample_(batch_size=batch_size, **kwargs)[0]

    @torch.no_grad()
    def sample_(
        self,
        batch_size: int = 1,
        preprocess_func: Callable | None = None
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Draws samples and their log probabilities from the model.

        Args:
            batch_size (int, optional): Number of samples to draw. Defaults to 1.
            preprocess_func (Callable, optional): A function to adjust the prior
                samples if needed. It should take samples and log probabilities as
                input and return modified values.

        Returns:
            torch.Tensor: Generated Samples (`y`)
            torch.Tensor: Log probabilities of the samples (`logq`)
        """
        x, logr = self._model.prior.sample_(batch_size)

        if preprocess_func is not None:
            x, logr = preprocess_func(x, logr)

        y, logj = self._model.net_(x)
        logq = logr - logj
        return y, logq

    @torch.no_grad()
    def sample_log(
        self,
        batch_size: int = 1,
        **kwargs
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Similar to `sample_`, but also returns the log probability of the
        target distribution from `model.action`.

        Args:
            batch_size (int, optional): Number of samples to draw. Defaults to 1.

        Returns:
            Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
                - `y`: Generated samples.
                - `logq`: Log probabilities of the samples.
                - `logp`: Log probabilities from the target distribution.
        """
        y, logq = self.sample_(batch_size=batch_size, **kwargs)
        logp = -self._model.action(y)  # logp is log(p_{non-normalized})
        return y, logq, logp

    @torch.no_grad()
    def log_prob(self, y: torch.Tensor) -> torch.Tensor:
        """Computes the log probability of the provided samples.

        Args:
            y (torch.Tensor): Samples for which to calculate the log probability.

        Returns:
            torch.Tensor: Log probabilities of the samples.
        """
        x, minus_logj = self._model.net_.reverse(y)
        logr = self._model.prior.log_prob(x)
        logq = logr + minus_logj
        return logq


class Fitter(TorchTrainer):
    """A class for training a given model."""
    def __init__(
        self,
        model: Model,
        epochs: int,
        config: Dict | TrainingConfiguration | None = None,
        strategy: Literal["ddp", "deepspeed", "horovod"] = 'ddp',
        logger: Logger | None = None,
        **kwargs
    ):
        super().__init__(
            config=config,
            epochs=epochs,
            strategy=strategy,
            logger=logger,
            **kwargs
        )
        self._model = model
        self.epochs = epochs
        # Global training configuration
        if isinstance(config, dict):
            config = TrainingConfiguration(**config)
        self.config = config
        self.checkpoint_dict = dict(
            display=self.config.ckpt_disp,
            print_stride=self.config.print_stride,
            print_batch_size=self.config.print_batch_size,
            snapshot_path=self.config.snapshot_path or 'checkpoint.pth',
            epochs_run=self.config.epochs_run
        )
        self.train_batch_size = 1
        self.train_history = dict(
            loss=[], logqp=[], logz=[], ess=[], rho=[], accept_rate=[]
        )
        self.hyperparam = dict(lr=self.config.optim_lr, weight_decay=self.config.weight_decay)

    def setup_seed(self, rank: int, world_size: int) -> None:
        """Sets up random seed for each worker in a distributed setting.
        This methods ensures that each workers receives a unique seed. The main worker
        generates the seeds and broadcasts to other workers.

        Args:
            rank (int): The rank of the current worker.
            world_size (int): The total number of workers in the distributed setting
        """
        if self.strategy.is_main_worker:
            # Generate unique seed for each worker based on its rank
            seeds_torch = [torch.randint(2**32 - 1, (1,)).item() for _ in range(world_size)]
            print(f"Generated seeds for workers: {seeds_torch}\n")
        else:
            seeds_torch = [None] * world_size

        # Broadcast seed list to all workers
        seeds_torch_list = self.strategy.allgather_obj(seeds_torch)
        seeds_torch = next(s for s in seeds_torch_list if s is not None)

        # Set seed for current worker based on its rank
        seed = seeds_torch[rank]
        torch.manual_seed(seed)

        print(f"Rank {rank} has been assigned seed {seed}\n")

    def create_dataloaders(
        self,
        train_dataset: Dataset,
        validation_dataset: Dataset | None = None,
        test_dataset: Dataset | None = None
    ) -> None:
        """Overrides create_dataloaders method from itwinai TorchTrainer and returns
        nothing since this is not needed in this use-case.
        """
        return

    def create_model_loss_optimizer(self) -> None:
        """Creates create_model_loss_optimizer method to setup model, seeds,
        loss function and optimizer.
        """
        self._model.prior.to(device=self.device)

        # First, define strategy-wise optional configurations
        if isinstance(self.strategy, DeepSpeedStrategy):
            # Batch size definition is not optional for DeepSpeedStrategy!
            distribute_kwargs = dict(
                config_params=dict(
                    train_micro_batch_size_per_gpu=self.config.batch_size
                )
            )
        else:
            distribute_kwargs = {}

        optimizer_class = eval(self.config.optimizer_class)
        if '_groups' in self._model.net_.__dict__.keys():
            parameters = self._model.net_.grouped_parameters()
        else:
            parameters = self._model.net_.parameters()
        self.optimizer = optimizer_class(parameters, **self.hyperparam)

        # Distributed model, optimizer, and scheduler
        self._model.net_, self.optimizer, _ = self.strategy.distributed(
            self._model.net_, self.optimizer, **distribute_kwargs
        )

        # Call setup_seed to set seed with current rank and world_size
        self.setup_seed(self.strategy.global_rank(), self.strategy.global_world_size())

        # decide whether to load a snapshot
        snapshot_path = self.checkpoint_dict['snapshot_path']
        if (snapshot_path is not None) and os.path.exists(snapshot_path):
            print(f"Trying to load snapshot from {snapshot_path}")
            self._load_snapshot()

        self.loss_fn = Fitter.calc_kl_mean if self.config.loss_fn == "None" else self.config.loss_fn

        if not self.config.scheduler or self.config.scheduler == "None":
            self.scheduler = None
        else:
            try:
                scheduler_class = getattr(lr_scheduler, self.config.scheduler)
                self.scheduler = scheduler_class(self.optimizer)
            except AttributeError:
                raise ValueError(f"Invalid scheduler name: {self.config.scheduler}")

    @suppress_workers_print
    def execute(
        self,
        train_dataset: Dataset | None = None,
        validation_dataset: Dataset | None = None,
        test_dataset: Dataset | None = None
    ) -> None:
        """Overrides execute method from itwinai TorchTrainer class to ignore
        train_dataset since it is not needed in this usecase.

        Args:
            train_dataset (Dataset | None): training dataset object. Defaults to None.
            validation_dataset (Dataset | None): validation dataset object.
                Defaults to None.
            test_dataset (Dataset | None): test dataset object. Defaults to None.
        """
        super().execute(train_dataset, validation_dataset, test_dataset)

    def _load_snapshot(self) -> None:
        """Method to load a snapshot from a path provided in checkpoint_dict"""
        snapshot_path = self.checkpoint_dict['snapshot_path']
        if torch.cuda.is_available():
            gpu_id = self.strategy.global_rank()
            loc = f"cuda:{gpu_id}"
            print(f"GPU: Attempting to load saved model into {loc}")
        else:
            loc = None  # cpu training
            print("CPU: Attempting to load saved model")
        snapshot = torch.load(snapshot_path, map_location=loc)
        self._model.net_.load_state_dict(snapshot["MODEL_STATE"])
        self.checkpoint_dict['epochs_run'] = snapshot['EPOCHS_RUN']
        print(f"Snapshot found: {snapshot_path}\nResuming training via Saved Snapshot at Epoch {snapshot['EPOCHS_RUN']}")

    def _save_snapshot(self, epoch: int) -> None:
        """Save snapshot of training for analysis and/or to continue training at a later date.

        Args:
            epoch (int): Number of the epoch at which the snapshot is saved.
        """
        snapshot_path = self.checkpoint_dict['snapshot_path']
        epochs_run = epoch + self.checkpoint_dict['epochs_run']
        snapshot_new_path = snapshot_path.rsplit('.',2)[0] + ".E" + str(epochs_run) + ".tar"
        snapshot = {
            "MODEL_STATE": self._model.net_.state_dict(),
            "EPOCHS_RUN": epochs_run
        }
        torch.save(snapshot, snapshot_new_path)
        print(f"Epoch {epochs_run} | Model Snapshot saved at {snapshot_new_path}")

    def set_epoch(self, epoch: int) -> None:
        """Sets current epoch at beginning of training.

        Args:
            epoch (int): current epoch number.
        """
        if self.profiler is not None:
            self.profiler.step()

    @profile_torch_trainer
    @measure_gpu_utilization
    def train(self) -> None:
        """Trains the neural network model."""

        # Track epoch time for scaling statistics
        if self.strategy.is_main_worker and self.strategy.is_distributed:
            try:
                num_nodes = int(os.environ["SLURM_NNODES"])
            except (KeyError, ValueError):
                logging.warning("SLURM_NNODES is not set or invalid; defaulting num_nodes to 1.")
                num_nodes = 1
            epoch_time_output_dir = Path("scalability-metrics/epoch-time")
            epoch_time_file_name = f"epochtime_{self.strategy.name}_{num_nodes}N.csv"
            epoch_time_output_path = epoch_time_output_dir / epoch_time_file_name

            epoch_time_logger= EpochTimeTracker(
                strategy_name=self.strategy.name,
                save_path=epoch_time_output_path,
                num_nodes=num_nodes,
                should_log=self.measure_epoch_time
            )

        if self.config.save_every == "None":
            self.config.save_every = self.epochs

        start_time = time.time()
        for epoch in range(1, self.epochs+1):
            self.set_epoch(epoch)
            loss, logqp = self.step(self.config.batch_size)
            self.checkpoint(epoch, loss, self.config.save_every)
            if self.scheduler is not None:
                self.scheduler.step()
        end_time = time.time()
        if self.strategy.is_main_worker:
            print(f"({self.strategy.device()}) Time = {end_time - start_time:.3g} sec.")

    def step(self, batch_size: int) -> None:
        """Perform a train step with a batch of inputs.

        Args:
            batch_size (int): Specifies the batch size for the training.
        """
        net_ = self._model.net_
        prior = self._model.prior
        action = self._model.action

        x, logr = prior.sample_(batch_size)
        y, logJ = net_(x)
        logq = logr - logJ
        logp = -action(y)
        loss = self.loss_fn(logq, logp)

        # clears old gradients from last steps
        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

        return loss, logq - logp

    def checkpoint(self, epoch: int, loss: torch.Tensor, save_every: int) -> None:
        """Handles model checkpointing and logging during training.

        Args:
            epoch (int): Epoch at which checkpoint is created.
            loss (torch.Tensor): Loss value at the epoch.
            save_every (int): interval at which checkpoints are saved.
        """
        print_stride = self.checkpoint_dict['print_stride']
        print_batch_size = self.checkpoint_dict['print_batch_size']
        snapshot_path = self.checkpoint_dict['snapshot_path']

        # Always save loss on rank 0
        if self.strategy.is_main_worker:
            self.train_history['loss'].append(loss.item())
            # Save model as well
            if snapshot_path is not None and (epoch % save_every == 0):
                self._save_snapshot(epoch)

        if epoch != 1 or epoch % print_stride != 0:
            return

        print_batch_size = print_batch_size // self.strategy.global_world_size()
        _, logq, logp = self._model.posterior.sample_log(print_batch_size)
        logq = self.strategy.gather(logq)
        logp = self.strategy.gather(logp)

        if not self.strategy.is_main_worker:
            return

        logq = torch.cat(logq, dim=0)
        logp = torch.cat(logp, dim=0)
        loss_ = self.loss_fn(logq, logp)
        self._append_to_train_history(logq, logp)
        self.print_fit_status(epoch, loss=loss_)

    @staticmethod
    def calc_kl_mean(logq: torch.Tensor, logp: torch.Tensor) -> torch.Tensor:
        """Return Kullback-Leibler divergence estimated from logq and logp.

        Args:
            logq (torch.Tensor): Log probabilities from the approximate distribution `q`.
            logp (torch.Tensor): Log probabilities from the target distribution `p`.

        Returns:
            torch.Tensor: Estimated KL divergence.
        """
        return (logq - logp).mean()  # KL, assuming samples from q

    @staticmethod
    def calc_kl_var(logq: torch.Tensor, logp: torch.Tensor) -> torch.Tensor:
        """Compute variance of the Kullback-Leibler (KL) divergence.

        Args:
            logq (torch.Tensor): Log probabilities from the approximate distribution `q`.
            logp (torch.Tensor): Log probabilities from the target distribution `p`.

        Returns:
            torch.Tensor: Variance of the KL divergence.
        """
        return (logq - logp).var()

    @staticmethod
    def calc_corrcoef(logq: torch.Tensor, logp: torch.Tensor) -> torch.Tensor:
        """Compute the Pearson correlation coefficient between logq and logp.

        Args:
            logq (torch.Tensor): Log probabilities from the approximate distribution `q`.
            logp (torch.Tensor): Log probabilities from the target distribution `p`.

        Returns:
            torch.Tensor: Pearson correlation coefficient.
        """
        return torch.corrcoef(torch.stack([logq, logp]))[0, 1]

    @staticmethod
    def calc_direct_kl_mean(logq: torch.Tensor, logp: torch.Tensor) -> torch.Tensor:
        """Compute the direct KL divergence.

        Args:
            logq (torch.Tensor): Log probabilities from the approximate distribution `q`.
            logp (torch.Tensor): Log probabilities from the target distribution `p`.

        Returns:
            torch.Tensor: KL divergence
        """
        logpq = logp - logq
        logz = torch.logsumexp(logpq, dim=0) - np.log(logp.shape[0])
        logpq = logpq - logz  # p is now normalized
        p_by_q = torch.exp(logpq)
        return (p_by_q * logpq).mean()

    @staticmethod
    def calc_minus_logz(logq: torch.Tensor, logp: torch.Tensor) -> torch.Tensor:
        """Compute the negative log partition function.

        Args:
            logq (torch.Tensor): Log probabilities from the approximate distribution `q`.
            logp (torch.Tensor): Log probabilities from the target distribution `p`.

        Returns:
            torch.Tensor: Negative log partition function.
        """
        logz = torch.logsumexp(logp - logq, dim=0) - np.log(logp.shape[0])
        return -logz

    @staticmethod
    def calc_ess(logq: torch.Tensor, logp: torch.Tensor) -> torch.Tensor:
        """Return effective sample size (ESS).

        Args:
            logq (torch.Tensor): Log probabilities from the approximate distribution `q`.
            logp (torch.Tensor): Log probabilities from the target distribution `p`.

        Returns:
            torch.Tensor: Estimated effective sample size (ESS).
        """
        logqp = logq - logp
        log_ess = 2*torch.logsumexp(-logqp, dim=0) - torch.logsumexp(-2*logqp, dim=0)
        ess = torch.exp(log_ess) / len(logqp)  # normalized
        return ess

    @staticmethod
    def calc_minus_logess(logq: torch.Tensor, logp: torch.Tensor) -> torch.Tensor:
        """Return logarithm of inverse of effective sample size.

        Args:
            logq (torch.Tensor): Log probabilities from the approximate distribution `q`.
            logp (torch.Tensor): Log probabilities from the target distribution `p`.

        Returns:
            torch.Tensor: logarithm of inverse of effective sample size.
        """
        logqp = logq - logp
        log_ess = 2*torch.logsumexp(-logqp, dim=0) - torch.logsumexp(-2*logqp, dim=0)
        return - log_ess + np.log(len(logqp))  # normalized

    @torch.no_grad()
    def _append_to_train_history(self, logq: torch.Tensor, logp: torch.Tensor) -> None:
        """Update train history with log probability estimates.

        Args:
            logq (torch.Tensor): Log probabilities from the approximate distribution `q`.
            logp (torch.Tensor): Log probabilities from the target distribution `p`.
        """
        logqp = logq - logp
        logz = estimate_logz(logqp, method='jackknife')  # returns (mean, std)
        accept_rate = self._model.mcmc.estimate_accept_rate(logqp)
        ess = self.calc_ess(logqp, 0)
        rho = self.calc_corrcoef(logq, logp)
        logqp = (logqp.mean().item(), logqp.std().item())
        self.train_history['logqp'].append(logqp)
        self.train_history['logz'].append(logz)
        self.train_history['ess'].append(ess)
        self.train_history['rho'].append(rho)
        self.train_history['accept_rate'].append(accept_rate)

    def print_fit_status(self, epoch: int, loss: torch.Tensor | None = None) -> None:
        """Print training progress and key metrics.

        Args:
            epoch (int): Current epoch.
            loss (torch.Tensor, optional): Loss value for the current epoch.
        """
        mydict = self.train_history
        if loss is None:
            loss = mydict['loss'][-1]
        logqp_mean, logqp_std = mydict['logqp'][-1]
        logz_mean, logz_std = mydict['logz'][-1]
        accept_rate_mean, accept_rate_std = mydict['accept_rate'][-1]
        # We now incorporate the effect of estimated log(z) to mean of log(q/p)
        adjusted_logqp_mean = logqp_mean + logz_mean
        ess = mydict['ess'][-1]
        rho = mydict['rho'][-1]

        if epoch == 1:
            print(f"\n>>> Training progress ({self.device}) <<<\n")
            print("Note: log(q/p) is estimated with normalized p; " \
                  + "mean & error are obtained from samples in a batch\n")

        epoch += self.checkpoint_dict['epochs_run']
        print(f"Epoch: {epoch} | loss: {loss:.4f} | ess: {ess:.4f}")
        self.log(loss.detach().cpu().numpy(),'epoch_loss',kind='metric')


# =============================================================================
@torch.no_grad()
def reverse_flow_sanitychecker(
    model: Model,
    n_samples: int = 4,
    net_: torch.nn.Module | None = None
) -> None:
    """Performs a sanity check on the reverse method of modules.
    Args:
        model (Model): Model containing prior and transformation networks.
        n_samples (int, optional): Number of samples to test. Defaults to 4.
        net_ (torch.nn.Module, optional): The transformation network.
    """
    if net_ is None:
        net_ = model.net_

    x = model.prior.sample(n_samples)
    y, logj = net_(x)
    x_hat, minus_logj = net_.backward(y)

    mean = lambda z: z.abs().mean().item()

    print("reverse method is OK if following values vanish (up to round off):")
    print(f"{mean(x - x_hat):g} & {mean(1 + minus_logj / logj):g}")



from typing import Dict, List, Tuple
import xarray as xr
from itwinai.components import DataSplitter, monitor_exec

from hython.scaler import Scaler
from hython.datasets import get_dataset
from hython.datasets.wflow_sbm import WflowSBM
from hython.config import Config


class RNNDatasetGetterAndPreprocessor(DataSplitter):
    def __init__(
        self,
        # == common ==
        hython_trainer: str,
        dataset: str,
        data_lazy_load: bool,
        scaling_variant: str,
        scaling_use_cached: bool,
        experiment_name: str,
        experiment_run: str,
        data_source: dict,
        work_dir: str,
        dynamic_inputs: List[str] | None = None,
        static_inputs: List[str] | None = None,
        target_variables: List[str] | None = None,
        scaling_static_range: Dict | None = None,
        mask_variables: List[str] | None = None,
        static_inputs_mask: List[str] | None = None,
        head_model_inputs: List[str] | None = None,
        train_temporal_range: List[str] = None,
        valid_temporal_range: List[str] = None,
        train_downsampler: Dict | None = None,
        valid_downsampler: Dict | None = None,
        downsampling_temporal_dynamic: bool | None = None,
        min_sample_target: int | None = None,
        seq_length: int | None = None
    ) -> None:
        self.save_parameters(**self.locals2params(locals()))

    @monitor_exec
    def execute(self) -> Tuple[WflowSBM, WflowSBM, None]:
        cfg = Config()

        for i in self.parameters:
            setattr(cfg, i, self.parameters[i])

        scaler = Scaler(cfg, cfg.scaling_use_cached)

        train_dataset = get_dataset(cfg.dataset)(cfg, scaler, True, "train")

        val_dataset = get_dataset(cfg.dataset)(cfg, scaler, False, "valid")

        return train_dataset, val_dataset, None



# Copyright (c) 2021-2023 Javad Komijani


from setuptools import setup


def readme():
    with open('README.rst') as f:
        return f.read()


packages = [
        'normflow',
        'normflow.action',
        'normflow.lib',
        'normflow.lib.combo',
        'normflow.lib.indexing',
        'normflow.lib.linalg',
        'normflow.lib.spline',
        'normflow.lib.stats',
        'normflow.mask',
        'normflow.mcmc',
        'normflow.nn',
        'normflow.nn.scalar',
        'normflow.prior'
        ]

package_dir = {
        'normflow': 'src',
        'normflow.action': 'src/action',
        'normflow.lib': 'src/lib',
        'normflow.lib.combo': 'src/lib/combo',
        'normflow.lib.indexing': 'src/lib/indexing',
        'normflow.lib.linalg': 'src/lib/linalg',
        'normflow.lib.spline': 'src/lib/spline',
        'normflow.lib.stats': 'src/lib/stats',
        'normflow.mask': 'src/mask',
        'normflow.mcmc': 'src/mcmc',
        'normflow.nn': 'src/nn',
        'normflow.nn.scalar': 'src/nn/scalar',
        'normflow.prior': 'src/prior'
        }

setup(name='normflow',
      version='1.1',
      description='Normalizing flow for generating lattice field configurations',
      packages=packages,
      package_dir=package_dir,
      url='http://github.com/jkomijani/normflow',
      author='Javad Komijani',
      author_email='jkomijani@gmail.com',
      license='MIT',
      install_requires=['numpy>=1.20', 'torch>=2.0'],
      zip_safe=False
      )



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Alex Krochak
#
# Credit:
# - Alex Krochak <o.krochak@fz-juelich.de> - FZJ
# --------------------------------------------------------------------------------------

import os
import torch
import glob
import numpy as np
import matplotlib.pyplot as plt

from itwinai.components import DataGetter, DataSplitter, monitor_exec
from torch.utils.data import Dataset, random_split

from typing import Optional, Tuple, Literal
from collections import OrderedDict
from pulsar_analysis.train_neural_network_model import ImageMaskPair, SignalLabelPair
from pulsar_analysis.preprocessing import PrepareFreqTimeImage, BinarizeToMask
from pulsar_analysis.pipeline_methods import (
    PipelineImageToMask,
    PipelineImageToFilterDelGraphtoIsPulsar,
    PipelineImageToFilterToCCtoLabels,
)
from pulsar_analysis.neural_network_models import UNet
from pulsar_simulation.generate_data_pipeline import generate_example_payloads_for_training

class SynthesizeData(DataGetter):
    def __init__(
        self,
        name: Optional[str] = None,
        tag: str = "test_v0_",
        num_payloads: int = 50,
        plot: bool = False,
        num_cpus: int = 4,
        param_root: str = "./syn_runtime/",
        payload_root: str = "./syn_payload/",
    ) -> None:
        """Initialize the synthesizeData class.
        Args:
            name [optional] (str):  name of the data getter component.
            param_root      (str):  folder where synthetic param data will be saved.
            payload_root    (str):  folder where synthetic payload data will be saved.
            tag             (str):  tag which is used as prefix for the generated files.
            num_cpus        (int):  number of CPUs used for parallel processing.
            num_payloads    (int):  number of generated examples.
            plot            (bool): if True, plotting routine is activated \
                                    (set False when running 'main.py' directly after)
            returns None.
        """
        super().__init__(name)
        self.save_parameters(**self.locals2params(locals()), pop_self=False)
        # TODO find a smart way to compute the right value for num_cpus
        os.makedirs(param_root, exist_ok=True)
        os.makedirs(payload_root, exist_ok=True)

    @monitor_exec
    def execute(self) -> None:
        """Generate synthetic data and save it to disk.
        Relies on the pulsar_simulation package."""
        generate_example_payloads_for_training(
            tag=self.parameters["tag"],
            num_payloads=self.parameters["num_payloads"],
            plot_a_example=self.parameters["plot"],
            param_folder=self.parameters["param_root"],
            payload_folder=self.parameters["payload_root"],
            num_cpus=self.parameters["num_cpus"],
            reinit_ray=False,
        )


class PulsarDataset(Dataset):
    """Class to represent common datasets. Variable 'engine_settings' is supposed to
    provide the settings for the image, mask and mask_maker engines, depending on
    the type of dataset. For example, for UNet dataset, the settings could be:

    engine_settings = {
        "image": {
            "do_rot_phase_avg": True,
            "do_binarize": False,
            "do_resize": True,
            "resize_size": (128, 128),
        },
        "mask": {
            "do_rot_phase_avg": True,
            "do_binarize": True,
            "do_resize": True,
            "resize_size": (128, 128),
        },
    }

    These settings are passed to the PrepareFreqTimeImage classes. The number and configuration
    of these are dependent on network type, thus initialization is inherently dependent on
    the provided "type" argument ! To clarify:
    - UNet: image and mask engines are initialized
        see: ImageToMaskDataset in src.pulsar_analysis.train_neural_network_model.py:150
    - FilterCNN: image, mask and mask_maker engines are initialized
        see: InMaskToMaskDataset in src.pulsar_analysis.train_neural_network_model.py:238
    - CNN1D: only mask engine is initialized.
        see: SignalToLabelDataset in src.pulsar_analysis.train_neural_network_model.py:496
    """

    def __init__(
        self,
        type: Literal["unet", "filtercnn", "cnn1d"],
        image_tag: str,
        mask_tag: str,
        image_directory: str,
        mask_directory: str,
        engine_settings: dict,
    ):
        self._type = type
        self._image_tag = image_tag
        self._mask_tag = mask_tag
        self._image_directory = image_directory
        self._mask_directory = mask_directory
        self._device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        self._image_engine = None
        self._mask_engine = None
        self._mask_maker_engine = None

        ### Compute dataset length ###
        search_pattern = os.path.join(self._image_directory, self._image_tag)
        matching_files = glob.glob(search_pattern)
        self._len = len(matching_files)

        ### Check that the engine settings are appropriate for chosen dataset type ###
        if self._type == "unet" and engine_settings.keys() != {"image","mask"}:
            raise ValueError("Wrong engine settings for UNet dataset. \n"
                "Provide 'image' and 'mask' engine settings.")

        elif self._type == "filtercnn" and \
        engine_settings.keys() != {"image","mask","mask_maker"}:
            raise ValueError("Wrong engine settings for FilterCNN dataset. \n"
            "Provide 'image', 'mask' and 'mask_maker' engine settings.")
        
        elif self._type == "cnn1d" and engine_settings.keys() != {"mask"}:
            raise ValueError("Wrong engine settings for CNN1D dataset.\n"
                             "Provide 'mask' engine settings.")

        ### Initialize the engines, one to three dependent on the dataset type ###
        ###        Mask engine settings are needed for all dataset types       ###

        # retrieve the binarization function from the engine settings and remove it
        bin_func = engine_settings["mask"].pop("binarize_func")
        bin_eng = BinarizeToMask(binarize_func=bin_func)

        self._mask_engine = PrepareFreqTimeImage(
            **engine_settings["mask"], binarize_engine=bin_eng
        )

        if self._type == "unet" or self._type == "filtercnn":  # init image engine
            self._image_engine = PrepareFreqTimeImage(**engine_settings["image"])

        if self._type == "filtercnn":  # init mask_maker engine
            # similarly, to bin. func., retrieve model type
            model_name = engine_settings["mask_maker"].pop("model")
            if model_name == "UNet":
                mme_model = UNet()
            else:
                raise ValueError(
                    "Unknown model type in engine_settings['mask_maker']['model']"
                )
            self._mask_maker_engine = PipelineImageToMask(
                **engine_settings["mask_maker"], image_to_mask_network=mme_model
            )

    def load_image_pair(self, img_id: int) -> ImageMaskPair | Tuple[torch.tensor, torch.tensor]:
        """Load an a data point from disk. Loading method depends on the network type:
        - For UNet and FilterCNN architectures, data point consits of image and mask pair.
        - For CNN1D architecture, data point consists of a mask only.
        """
        img_address = self._image_directory + self._image_tag.replace("*", str(img_id))
        mask_address = self._mask_directory + self._mask_tag.replace("*", str(img_id))

        if self._type == "unet":
            pair = ImageMaskPair.load_from_payload_address(
                image_payload_address=img_address,
                mask_payload_address=mask_address,
                image_engine=self._image_engine,
                mask_engine=self._mask_engine,
            )

        elif self._type == "filtercnn":
            pair = ImageMaskPair.load_from_payload_and_make_in_mask(
                image_payload_address=img_address,
                mask_payload_address=mask_address,
                mask_maker_engine=self._mask_maker_engine,
                image_engine=self._image_engine,
                mask_engine=self._mask_engine,
            )

        elif self._type == "cnn1d":
            signal_label_pair = SignalLabelPair.load_from_payload_address(
                mask_payload_address=mask_address, mask_engine=self._mask_engine
            )

            signal, label = signal_label_pair()
            label_vector = list(label.values())
            pulsar_present = label_vector[0]

            if pulsar_present > 0:
                label_vector = [1]
            else:
                label_vector = [0]

            label_vector = torch.tensor(label_vector, dtype=torch.float32)
            pair = torch.tensor(signal, dtype=torch.float32).unsqueeze(0), label_vector

        return pair

    def __getitem__(self, item_id: int) -> ImageMaskPair | Tuple[torch.tensor, torch.tensor]:
        """Use load_image_pair method to retrieve the data point with the given item_id.
        Output is network architecture dependent.
        Returns either image-mask pair (UNET, FCNN) or signal-label pair (CNN1D)."""

        ### Implementation is slightly different per architecture
        ### due to the original use-case code
        if self._type == "cnn1d":
            return self.load_image_pair(item_id)
        elif self._type == "unet" or self._type == "filtercnn":
            img, mask = self.load_image_pair(item_id)()
            img = img.unsqueeze(0)
            mask = mask.unsqueeze(0)
            return img.float(), mask.float()

    def __get_descriptions__(self, item_id: int):
        """Provide descriptions of the data point with the given item_id."""
        return self.load_image_pair(item_id).descriptions

    def __len__(self):
        """Return the length of the dataset pre-computed during the initialization."""
        return self._len

    def plot(self, item_id: int):
        """Plot the data point with the provided item_id.
        Plotting method is model-dependent."""
        ##TODO: make this method HPC-friendy
        if self._type == "unet" or self._type == "filtercnn":
            image_payload_address = self._image_directory + self._image_tag.replace(
                "*", str(item_id)
            )
            mask_payload_address = self._mask_directory + self._mask_tag.replace(
                "*", str(item_id)
            )

            image_mask_pair = ImageMaskPair.load_from_payload_and_make_in_mask(
                image_payload_address=image_payload_address,
                mask_payload_address=mask_payload_address,
                mask_maker_engine=self._mask_maker_engine,
                image_engine=self._image_engine,
                mask_engine=self._mask_engine,
            )
            image_mask_pair.plot()

        elif self._type == "cnn1d":
            signal, label_vector = self[item_id]
            _, ax = plt.subplots(1, 1, figsize=(3.5, 3.5))
            ax.plot(signal.detach().numpy().flatten())
            ax.set_xlabel("freqs (a.u)")
            ax.set_ylabel("lags")
            ax.set_aspect("auto")
            ax.set_ylim(0, 1)
            ax.grid()
            ax.set_title(f"Pulsar Present {label_vector[0]>=0.9}")

    def execute(self) -> Dataset:
        """Read the dataset from disk and return it to the trainer in-memory."""
        return self


class DatasetSplitter(DataSplitter):
    def __init__(
        self,
        train_proportion: int | float,
        validation_proportion: int | float = 0.0,
        test_proportion: int | float = 0.0,
        rnd_seed: Optional[int] = None,
        name: Optional[str] = None,
    ) -> None:
        """Initialize the splitter for time-series datasets.

        Args:
            train_proportion (int | float): Proportion of files for the training set.
            validation_proportion (int | float): Proportion for validation.
            test_proportion (int | float): Proportion for testing.
            rnd_seed (int | None): Seed for randomization (optional).
            name (str | None): Name of the data splitter.
        """
        super().__init__(train_proportion, validation_proportion, test_proportion, name)
        self.save_parameters(**self.locals2params(locals()))
        self.rnd_seed = rnd_seed

    @monitor_exec
    def execute(self, whole_dataset: Dataset) -> Tuple[Dataset, Dataset, Dataset]:
        """Execute the dataset splitting process.

        Finds all pickled files in the root folder, then splits them into
        training, validation, and test sets based on the specified proportions.

        Returns:
            Tuple[Dataset, Dataset, Dataset]: Training, validation, and test datasets.
        """

        # Split file paths into train, validation, and test sets
        generator = torch.Generator().manual_seed(self.rnd_seed)
        train_dataset, validation_dataset, test_dataset = random_split(
            whole_dataset,
            [self.train_proportion, self.validation_proportion, self.test_proportion],
            generator=generator,
        )
        return train_dataset, validation_dataset, test_dataset


class PipelinePulsarInterface(PipelineImageToFilterDelGraphtoIsPulsar):
    def execute(self) -> PipelineImageToFilterDelGraphtoIsPulsar:
        return self


class PipelineLabelsInterface(PipelineImageToFilterToCCtoLabels):
    def execute(self) -> PipelineImageToFilterToCCtoLabels:
        return self


class TestSuite:
    def __init__(
        self,
        image_to_mask_network: torch.nn.Module,
        trained_image_to_mask_network_path: str,
        mask_filter_network: torch.nn.Module,
        trained_mask_filter_network_path: str,
        signal_to_label_network: torch.nn.Module,
        trained_signal_to_label_network: str,
        img_dir: str,
        lbl_dir: str,
        size: int,
        offset: int,
    ):
        self.img_dir = img_dir
        self.lbl_dir = lbl_dir
        self.size = size
        self.offset = offset

        self.del_graph_to_is_pulsar = PipelineImageToFilterDelGraphtoIsPulsar(
            image_to_mask_network,
            trained_image_to_mask_network_path,
            mask_filter_network,
            trained_mask_filter_network_path,
            signal_to_label_network,
            trained_signal_to_label_network,
        )

        self.to_cc_to_labels = PipelineImageToFilterToCCtoLabels(
            image_to_mask_network,
            trained_image_to_mask_network_path,
            mask_filter_network,
            trained_mask_filter_network_path,
            min_cc_size_threshold=5,
        )

    def execute(self):
        data = np.load(file=self.img_dir, mmap_mode="r")
        data_label = np.load(file=self.lbl_dir, mmap_mode="r")
        data_subset = data[self.offset + 1 : self.offset + self.size, :, :]
        data_label_subset = data_label[self.offset + 1 : self.offset + self.size]

        self.del_graph_to_is_pulsar.test_on_real_data_from_npy_files(
            image_data_set=data_subset,
            image_label_set=data_label_subset,
            plot_details=True,
            plot_randomly=True,
            batch_size=2,
        )

        self.to_cc_to_labels.test_on_real_data_from_npy_files(
            image_data_set=data_subset,
            image_label_set=data_label_subset,
            plot_randomly=True,
            batch_size=2,
        )

        # plt.show()
        os.makedirs("plots", exist_ok=True)
        for i in plt.get_fignums():
            fig = plt.figure(i)
            fig.savefig(f"plots/figure_{i}.png")

        print("Test Suite executed")


class ModelSaver:
    def execute(self, model, path) -> None:
        m_dict = model.state_dict()
        # ensure correct saving syntax expected by the loader
        m_new = OrderedDict(
            [
                (k.replace("module.", ""), v) if k.startswith("module") else (k, v)
                for k, v in m_dict.items()
            ]
        )
        torch.save(m_new, path)
        print(f"Model saved at {path}")



import numpy as np
import matplotlib.pyplot as plt
import os

from pulsar_analysis.neural_network_models import (
    UNet,
    FilterCNN,
    CNN1D,
    WeightedBCELoss,
)

from pulsar_analysis.pipeline_methods import (
    PipelineImageToFilterDelGraphtoIsPulsar,
    PipelineImageToFilterToCCtoLabels,
)

from pulsar_simulation.generate_data_pipeline import generate_example_payloads_for_training

# itwinai integration
from trainer import PulsarTrainer
from data import PulsarDataset, DatasetSplitter
import inputs

if inputs.generate_data:
    os.makedirs(inputs.param_folder, exist_ok=True)
    os.makedirs(inputs.payload_folder, exist_ok=True)

    generate_example_payloads_for_training(
        tag=inputs.tag,
        num_payloads=inputs.num_payloads,
        plot_a_example=inputs.plot_a_example,
        param_folder=inputs.param_folder,
        payload_folder=inputs.payload_folder,
        num_cpus=inputs.num_cpus,
    )

### INITIALIZE THE DATASETS ###

unet_ds = PulsarDataset(
    image_tag=inputs.image_tag,
    mask_tag=inputs.mask_tag,
    image_directory=inputs.image_directory,
    mask_directory=inputs.mask_directory,
    type="unet",
    engine_settings=inputs.engine_settings_unet,
)

fcnn_ds = PulsarDataset(
    image_tag=inputs.image_tag,
    mask_tag=inputs.mask_tag,
    image_directory=inputs.image_directory,
    mask_directory=inputs.mask_directory,
    type="filtercnn",
    engine_settings=inputs.engine_settings_filtercnn,
)

cnn1d_ds = PulsarDataset(
    image_tag=inputs.image_tag,
    mask_tag=inputs.mask_tag,
    image_directory=inputs.image_directory,
    mask_directory=inputs.mask_directory,
    type="cnn1d",
    engine_settings=inputs.engine_settings_cnn1d,
)

### INITIALIZE THE SPLITTER ###

splitter = DatasetSplitter(
    train_proportion=inputs.train_proportion,
    validation_proportion=inputs.validation_proportion,
    test_proportion=inputs.test_proportion,
    rnd_seed=inputs.rnd_seed,
    name=inputs.spl_name,
)

unet_trainer = PulsarTrainer(
    model=UNet(),
    epochs=1,
    store_trained_model_at=inputs.unet_dir,
    loss=WeightedBCELoss(pos_weight=3, neg_weight=1),
    name="UNet Trainer",
    config=inputs.config,
)

fcnn_trainer = PulsarTrainer(
    model=FilterCNN(),
    epochs=1,
    store_trained_model_at=inputs.fcnn_dir,
    loss=WeightedBCELoss(pos_weight=1, neg_weight=1),
    name="FilterCNN Trainer",
    config=inputs.config,
)

cnn1d_trainer = PulsarTrainer(
    model=CNN1D(),
    epochs=1,
    loss=WeightedBCELoss(pos_weight=1, neg_weight=1),
    store_trained_model_at=inputs.cnn1d_dir,
    name="SignalToLabelTrainer",
    config=inputs.config,
)

unet_trainer.execute(*splitter.execute(unet_ds))
fcnn_trainer.execute(*splitter.execute(fcnn_ds))
cnn1d_trainer.execute(*splitter.execute(cnn1d_ds))

unet_trainer.write_model()
fcnn_trainer.write_model()
cnn1d_trainer.write_model()


if input("Evaluate model on real data? [y/N]: ") == "y":
    ## assemble the full pipeline and test on real data
    ppl1f = PipelineImageToFilterDelGraphtoIsPulsar(
        image_to_mask_network=UNet(),
        trained_image_to_mask_network_path=inputs.unet_dir,
        mask_filter_network=FilterCNN(),
        trained_mask_filter_network_path=inputs.fcnn_dir,
        signal_to_label_network=CNN1D(),
        trained_signal_to_label_network=inputs.cnn1d_dir,
    )

    ppl2f = PipelineImageToFilterToCCtoLabels(
        image_to_mask_network=UNet(),
        trained_image_to_mask_network_path=inputs.unet_dir,
        mask_filter_network=FilterCNN(),
        trained_mask_filter_network_path=inputs.fcnn_dir,
        min_cc_size_threshold=5,
    )

    # Load and format the evaluation data
    data = np.load(file=inputs.image_directory_npy, mmap_mode="r")
    data_label = np.load(file=inputs.label_directory_npy, mmap_mode="r")
    data_subset = data[inputs.offset + 1 : inputs.offset + inputs.size_of_set, :, :]
    data_label_subset = data_label[inputs.offset + 1 : inputs.offset + inputs.size_of_set]

    ppl1f.test_on_real_data_from_npy_files(
        image_data_set=data_subset,
        image_label_set=data_label_subset,
        plot_details=True,
        plot_randomly=True,
        batch_size=2,
    )
    ppl2f.test_on_real_data_from_npy_files(
        image_data_set=data_subset,
        image_label_set=data_label_subset,
        plot_randomly=True,
        batch_size=2,
    )
    plt.show()
    plt.close()



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# - Anna Lappe <anna.elisa.lappe@cern.ch> - CERN
#
# --------------------------------------------------------------------------------------
# Command-line interface for the itwinai Python library.
# Example:
#
# >>> itwinai --help
#
# --------------------------------------------------------------------------------------
#
# NOTE: import libraries in the command's function, not here, as having them here will
# slow down the CLI commands significantly.

import logging
import os
import subprocess
import sys
from pathlib import Path
from typing import Dict, List, Optional

import hydra
import typer
from typing_extensions import Annotated

app = typer.Typer(pretty_exceptions_enable=False)

py_logger = logging.getLogger(__name__)


@app.command()
def generate_flamegraph(
    file: Annotated[str, typer.Option(help="The location of the raw profiling data.")],
    output_filename: Annotated[
        str, typer.Option(help="The filename of the resulting flamegraph.")
    ] = "flamegraph.svg",
):
    """Generates a flamegraph from the given profiling output."""
    script_filename = "flamegraph.pl"
    script_path = Path(__file__).parent / script_filename

    if not script_path.exists():
        py_logger.exception(f"Could not find '{script_filename}' at '{script_path}'")
        raise typer.Exit()

    try:
        with open(output_filename, "w") as out:
            subprocess.run(
                ["perl", str(script_path), file],
                stdout=out,
                check=True,
            )
        typer.echo(f"Flamegraph saved to '{output_filename}'")
    except FileNotFoundError:
        typer.echo("Error: Perl is not installed or not in PATH.")
    except subprocess.CalledProcessError as e:
        typer.echo(f"Flamegraph generation failed: {e}")


@app.command()
def generate_py_spy_report(
    file: Annotated[str, typer.Option(help="The location of the raw profiling data.")],
    num_rows: Annotated[
        str,
        typer.Option(help="Number of rows to display. Pass 'all' to print the full table."),
    ] = "10",
    aggregate_leaf_paths: Annotated[
        bool,
        typer.Option(
            help="Whether to aggregate all unique leaf calls across different call stacks."
        ),
    ] = False,
    library_name: Annotated[
        str, typer.Option(help="Which library name to find the lowest contact point of.")
    ] = "itwinai",
):
    """Generates a short aggregation of the raw py-spy profiling data, showing which leaf
    functions collected the most samples.
    """
    from tabulate import tabulate

    from itwinai.torch.profiling.py_spy_aggregation import (
        add_lowest_library_function,
        convert_stack_trace_to_list,
        get_aggregated_paths,
    )

    if not num_rows.isnumeric() and num_rows != "all":
        raise typer.BadParameter(
            f"Number of rows must be either an integer or 'all'. Was '{num_rows}'.",
            param_hint="num-rows",
        )
    parsed_num_rows: int | None = int(num_rows) if num_rows.isnumeric() else None
    if isinstance(parsed_num_rows, int) and parsed_num_rows < 1:
        raise typer.BadParameter(
            f"Number of rows must be at least one! Was '{num_rows}'.",
            param_hint="num-rows",
        )

    file_path = Path(file)
    if not file_path.exists():
        raise typer.BadParameter(f"'{file_path.resolve()}' was not found!", param_hint="file")

    # Reading and converting the data
    with file_path.open("r") as f:
        profiling_data = f.readlines()

    stack_traces: List[List[Dict]] = []
    for line in profiling_data:
        try:
            structured_stack_trace = convert_stack_trace_to_list(line)
            if structured_stack_trace:
                stack_traces.append(structured_stack_trace)
        except ValueError as exception:
            typer.echo(f"Failed to aggregate data with following error:\n{exception}")
            raise typer.Exit()

    add_lowest_library_function(stack_traces=stack_traces, library_name=library_name)
    leaf_functions = [data_point[-1] for data_point in stack_traces]
    if aggregate_leaf_paths:
        leaf_functions = get_aggregated_paths(functions=leaf_functions)

    leaf_functions.sort(key=lambda x: x["num_samples"], reverse=True)

    # Turn num_samples into percentages
    total_samples = sum(function_dict["num_samples"] for function_dict in leaf_functions)
    for function_dict in leaf_functions:
        num_samples = function_dict["num_samples"]
        percentage = 100 * num_samples / total_samples
        function_dict["proportion (n)"] = f"{percentage:.2f}% ({num_samples})"
        del function_dict["num_samples"]

    typer.echo(tabulate(leaf_functions[:parsed_num_rows], headers="keys", tablefmt="presto"))


@app.command()
def generate_scalability_report(
    log_dir: Annotated[
        str,
        typer.Option(help=("Which directory to search for the scalability metrics in.")),
    ] = "scalability-metrics",
    plot_dir: Annotated[
        str, typer.Option(help=("Which directory to save the resulting plots in."))
    ] = "plots",
    do_backup: Annotated[
        bool,
        typer.Option(
            help=(
                "Whether to store a backup of the scalability metrics that were used"
                " to make the report or not."
            )
        ),
    ] = False,
    run_ids: Annotated[
        str | None,
        typer.Option(
            help=(
                "Which run ids to read, presented as comma-separated values, e.g. 'run0,run1'."
            )
        ),
    ] = None,
    backup_root_dir: Annotated[
        str, typer.Option(help=("Which directory to store the backup files in."))
    ] = "backup-scalability-metrics/",
    plot_file_suffix: Annotated[
        str,
        typer.Option(
            help=(
                "Which file suffix to use for the plots. Useful for changing between raster"
                " and vector based images"
            )
        ),
    ] = ".png",
):
    """Generates scalability reports for epoch time, GPU data, and communication data
    based on log files in the specified directory. Optionally, backups of the reports
    can be created.

    This command processes log files stored in specific subdirectories under the given
    `log_dir`. It generates plots and metrics for scalability analysis and saves them
    in the `plot_dir`. If backups are enabled, the generated reports will also be
    copied to a backup directory under `backup_root_dir`.
    """
    from datetime import datetime

    from itwinai.scalability_report.reports import (
        communication_data_report,
        epoch_time_report,
        gpu_data_report,
    )

    log_dir_path = Path(log_dir)
    if not log_dir_path.exists():
        raise ValueError(f"The provided log_dir, '{log_dir_path.resolve()}', does not exist.")

    if run_ids:
        base_directories_for_runs = [log_dir_path / run_id for run_id in run_ids.split(",")]
        # Ensure that all passed run_ids actually exist as directories
        non_existent_paths = [
            str(path.resolve()) for path in base_directories_for_runs if not path.exists()
        ]
        if non_existent_paths:
            raise ValueError(f"Given run_id paths do not exist: '{non_existent_paths}'!")
    else:
        # Ensure that all elements in log_dir are directories
        non_directory_paths = [
            str(path.resolve()) for path in log_dir_path.iterdir() if not path.is_dir()
        ]
        if non_directory_paths:
            raise ValueError(
                f"Found elements in log_dir that are not directories: '{non_directory_paths}'"
            )
        base_directories_for_runs = list(log_dir_path.iterdir())

    # Finding the respective data logging directories
    epoch_time_logdirs = [
        path / "epoch-time"
        for path in base_directories_for_runs
        if (path / "epoch-time").exists()
    ]
    gpu_data_logdirs = [
        path / "gpu-energy-data"
        for path in base_directories_for_runs
        if (path / "gpu-energy-data").exists()
    ]
    comm_time_logdirs = [
        path / "communication-data"
        for path in base_directories_for_runs
        if (path / "communication-data").exists()
    ]

    # Setting the backup directory from run name
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    if base_directories_for_runs:
        backup_run_id = "_".join(map(str, base_directories_for_runs)) + f"_{timestamp}"
    else:
        backup_run_id = f"aggregated_run_{timestamp}"
    backup_dir = Path(backup_root_dir) / backup_run_id

    epoch_time_backup_dir = backup_dir / "epoch-time"
    gpu_data_backup_dir = backup_dir / "gpu-energy-data"
    communication_data_backup_dir = backup_dir / "communication-data"

    plot_dir_path = Path(plot_dir)
    plot_dir_path.mkdir(exist_ok=True, parents=True)

    epoch_time_table = epoch_time_report(
        log_dirs=epoch_time_logdirs,
        plot_dir=plot_dir_path,
        backup_dir=epoch_time_backup_dir,
        do_backup=do_backup,
        plot_file_suffix=plot_file_suffix,
    )
    gpu_data_table = gpu_data_report(
        log_dirs=gpu_data_logdirs,
        plot_dir=plot_dir_path,
        backup_dir=gpu_data_backup_dir,
        do_backup=do_backup,
        plot_file_suffix=plot_file_suffix,
    )
    communication_data_table = communication_data_report(
        log_dirs=comm_time_logdirs,
        plot_dir=plot_dir_path,
        backup_dir=communication_data_backup_dir,
        do_backup=do_backup,
        plot_file_suffix=plot_file_suffix,
    )

    typer.echo("")
    if epoch_time_table is not None:
        typer.echo("#" * 8 + " Epoch Time Report " + "#" * 8)
        typer.echo(epoch_time_table + "\n")
    else:
        typer.echo("No Epoch Time Data Found\n")

    if gpu_data_table is not None:
        typer.echo("#" * 8 + "GPU Data Report" + "#" * 8)
        typer.echo(gpu_data_table + "\n")
    else:
        typer.echo("No GPU Data Found\n")

    if communication_data_table is not None:
        typer.echo("#" * 8 + "Communication Data Report" + "#" * 8)
        typer.echo(communication_data_table, "\n")
    else:
        typer.echo("No Communication Data Found\n")


@app.command()
def sanity_check(
    torch: Annotated[
        Optional[bool], typer.Option(help=("Check also itwinai.torch modules."))
    ] = False,
    tensorflow: Annotated[
        Optional[bool], typer.Option(help=("Check also itwinai.tensorflow modules."))
    ] = False,
    all: Annotated[Optional[bool], typer.Option(help=("Check all modules."))] = False,
    optional_deps: List[str] = typer.Option(None, help="List of optional dependencies."),
):
    """Run sanity checks on the installation of itwinai and its dependencies by trying
    to import itwinai modules. By default, only itwinai core modules (neither torch, nor
    tensorflow) are tested."""
    from itwinai.tests.sanity_check import (
        run_sanity_check,
        sanity_check_all,
        sanity_check_slim,
        sanity_check_tensorflow,
        sanity_check_torch,
    )

    all = (torch and tensorflow) or all
    if all:
        sanity_check_all()
    elif torch:
        sanity_check_torch()
    elif tensorflow:
        sanity_check_tensorflow()
    else:
        sanity_check_slim()

    if optional_deps is not None:
        run_sanity_check(optional_deps)


@app.command(context_settings={"allow_extra_args": True, "ignore_unknown_options": True})
def generate_slurm(
    job_name: Annotated[
        str | None, typer.Option("--job-name", help="The name of the SLURM job.")
    ] = None,
    account: Annotated[
        str, typer.Option("--account", help="The billing account for the SLURM job.")
    ] = "intertwin",
    time: Annotated[
        str, typer.Option("--time", help="The time limit of the SLURM job.")
    ] = "00:30:00",
    partition: Annotated[
        str,
        typer.Option(
            "--partition",
            help="Which partition of the cluster the SLURM job is going to run on.",
        ),
    ] = "develbooster",
    std_out: Annotated[
        str | None, typer.Option("--std-out", help="The standard out file.")
    ] = None,
    err_out: Annotated[
        str | None, typer.Option("--err-out", help="The error out file.")
    ] = None,
    num_nodes: Annotated[
        int,
        typer.Option(
            "--num-nodes",
            help="The number of nodes that the SLURM job is going to run on.",
        ),
    ] = 1,
    num_tasks_per_node: Annotated[
        int, typer.Option("--num-tasks-per-node", help="The number of tasks per node.")
    ] = 1,
    gpus_per_node: Annotated[
        int,
        typer.Option("--gpus-per-node", help="The requested number of GPUs per node."),
    ] = 4,
    cpus_per_gpu: Annotated[
        int,
        typer.Option("--cpus-per-gpu", help="The requested number of CPUs per GPU."),
    ] = 4,
    config_path: Annotated[
        str,
        typer.Option(
            "--config-path",
            help="The path to the directory containing the config file to use for training.",
        ),
    ] = ".",
    config_name: Annotated[
        str,
        typer.Option("--config-name", help="The name of the config file to use for training."),
    ] = "config",
    pipe_key: Annotated[
        str,
        typer.Option("--pipe-key", help="Which pipe key to use for running the pipeline."),
    ] = "rnn_training_pipeline",
    mode: Annotated[
        str,
        typer.Option(
            "--mode",
            help="Which mode to run, e.g. scaling test, all strategies, or a single run.",
            case_sensitive=False,
        ),
    ] = "single",
    dist_strat: Annotated[
        str,
        typer.Option(
            "--dist-strat",
            help="Which distributed strategy to use.",
            case_sensitive=False,
        ),
    ] = "ddp",
    pre_exec_cmd: Annotated[
        str | None,
        typer.Option(
            "--pre-exec-cmd", help="The pre-execution command to use for the python script."
        ),
    ] = None,
    training_cmd: Annotated[
        str | None,
        typer.Option(
            "--training-cmd", help="The training command to use for the python script."
        ),
    ] = None,
    python_venv: Annotated[
        str,
        typer.Option(
            "--python-venv", help="Which python venv to use for running the command."
        ),
    ] = ".venv",
    scalability_nodes: Annotated[
        str,
        typer.Option(
            "--scalability-nodes",
            help="A comma-separated list of node numbers to use for the scalability test.",
        ),
    ] = "1,2,4,8",
    debug: Annotated[
        bool,
        typer.Option("--debug", help="Whether to include debugging information or not"),
    ] = False,
    no_save_script: Annotated[
        bool,
        typer.Option(
            "--no-save-script", help="Whether to save the script after processing it."
        ),
    ] = False,
    no_submit_job: Annotated[
        bool,
        typer.Option(
            "--no-submit-job",
            help="Whether to submit the job when processing the script.",
        ),
    ] = False,
    config: Annotated[
        str | None,
        typer.Option("--config", help="The path to the SLURM configuration file."),
    ] = None,
    py_spy: Annotated[
        bool, typer.Option("--py-spy", help="Whether to activate profiling with py-spy or not")
    ] = False,
    profiling_rate: Annotated[
        int,
        typer.Option(
            "--profiling-rate", help="The rate at which to profile with the py-spy profiler."
        ),
    ] = 10,
):
    """Generates a default SLURM script using arguments and optionally a configuration
    file.
    """
    from itwinai.slurm.slurm_script_builder import generate_default_slurm_script

    del sys.argv[0]
    generate_default_slurm_script()


@app.command(context_settings={"allow_extra_args": True, "ignore_unknown_options": True})
def exec_pipeline(
    # NOTE: The arguments below are not actually needed in this function, but they are here
    # to replicate Hydra's help page in Typer, making it easier for the users to use it.
    hydra_help: Annotated[bool, typer.Option(help="Show Hydra's help page")] = False,
    version: Annotated[bool, typer.Option(help="Show Hydra's version and exit")] = False,
    cfg: Annotated[
        str,
        typer.Option("--cfg", "-c", help="Show config instead of running [job|hydra|all]"),
    ] = "",
    resolve: Annotated[
        bool,
        typer.Option(
            help="Used in conjunction with --cfg, resolve "
            "config interpolations before printing."
        ),
    ] = False,
    package: Annotated[
        str,
        typer.Option("--package", "-p", help="Config package to show"),
    ] = "",
    run: Annotated[
        str,
        typer.Option("--run", "-r", help="Run a job"),
    ] = "",
    multirun: Annotated[
        str,
        typer.Option(
            "--multirun",
            "-m",
            help="Run multiple jobs with the configured launcher and sweeper",
        ),
    ] = "",
    shell_completion: Annotated[
        str,
        typer.Option(
            "--shell-completion",
            "-sc",
            help="Install or Uninstall shell completion",
        ),
    ] = "",
    config_path: Annotated[
        str,
        typer.Option(
            "--config-path",
            "-cp",
            help=(
                # NOTE: this docstring changed from Hydra's help page.
                "Overrides the config_path specified in hydra.main(). "
                "The config_path is absolute, or relative to the current workign directory. "
                "Defaults to the current working directory."
            ),
        ),
    ] = "",
    config_name: Annotated[
        str,
        typer.Option(
            "--config-name",
            "-cn",
            help="Overrides the config_name specified in hydra.main()",
        ),
    ] = "config",
    config_dir: Annotated[
        str,
        typer.Option(
            "--config-dir",
            "-cd",
            help="Adds an additional config dir to the config search path",
        ),
    ] = "",
    experimental_rerun: Annotated[
        str,
        typer.Option(
            "--experimental-rerun",
            help="Rerun a job from a previous config pickle",
        ),
    ] = "",
    info: Annotated[
        str,
        typer.Option(
            "--info",
            "-i",
            help=(
                "Print Hydra information "
                "[all|config|defaults|defaults-tree|plugins|searchpath]"
            ),
        ),
    ] = "",
    overrides: Annotated[
        Optional[List[str]],
        typer.Argument(
            help=(
                "Any key=value arguments to override config values "
                "(use dots for.nested=overrides), using the Hydra syntax."
            ),
        ),
    ] = None,
):
    """Execute a pipeline from configuration file using Hydra CLI. Allows dynamic override
    of fields which can be appended as a list of overrides (e.g., batch_size=32).
    By default, it will expect a configuration file called "config.yaml" in the
    current working directory. To override the default behavior set --config-name and
    --config-path.
    By default, this command will execute the whole pipeline under "training_pipeline"
    field in the configuration file. To execute a different pipeline you can override this
    by passing "+pipe_key=your_pipeline" in the list of overrides, and to execute only a
    subset of the steps, you can pass "+pipe_steps=[0,1]".
    """
    from itwinai.utils import make_config_paths_absolute

    del sys.argv[0]

    # Add current working directory to the module search path
    # so hydra will find the objects defined in the config (usually paths relative to config)
    sys.path.append(os.getcwd())

    # Process CLI arguments to handle paths
    sys.argv = make_config_paths_absolute(sys.argv)

    exec_pipeline_with_compose()


@hydra.main(version_base=None, config_path=os.getcwd(), config_name="config")
def exec_pipeline_with_compose(cfg):
    """Hydra entry function. The hydra.main decorator parses a configuration file
    (under config_path), which contains a pipeline definition, and passes it to this function
    as an omegaconf.DictConfig object (called cfg). This function then instantiates and
    executes the resulting pipeline object.
    Filters steps if `pipe_steps` is provided, otherwise executes the entire pipeline.
    For more information on hydra.main, please see
    https://hydra.cc/docs/tutorials/basic/your_first_app/simple_cli/."""

    from hydra.utils import instantiate
    from omegaconf import OmegaConf, errors

    # Register custom OmegaConf resolver to allow to dynaimcally compute the current working
    # directory. Example: some_field: ${itwinai.cwd:}/some/nested/path/in/current/working/dir
    OmegaConf.register_new_resolver("itwinai.cwd", lambda: os.getcwd())

    def range_resolver(x, y=None, step=1):
        """Custom OmegaConf resolver for range."""
        if y is None:
            return list(range(int(x)))
        return list(range(int(x), int(y), int(step)))

    # Register custom OmegaConf resolver to allow to dynaimcally compute ranges
    OmegaConf.register_new_resolver("itwinai.range", range_resolver)

    # Register custom OmegaConf resolver to allow to dynaimcally compute ranges
    OmegaConf.register_new_resolver("itwinai.multiply", lambda x, y: x * y)

    pipe_steps = OmegaConf.select(cfg, "pipe_steps", default=None)
    pipe_key = OmegaConf.select(cfg, "pipe_key", default="training_pipeline")

    try:
        cfg = OmegaConf.select(cfg, pipe_key, throw_on_missing=True)
    except errors.MissingMandatoryValue as e:
        e.add_note(
            f"Could not find pipeline key {pipe_key}. Make sure that you provide the full "
            "dotpath to your pipeline key."
        )
        raise e

    if pipe_steps:
        try:
            cfg.steps = [cfg.steps[step] for step in pipe_steps]
            typer.echo(f"Successfully selected steps {pipe_steps}")
        except errors.ConfigKeyError as e:
            e.add_note(
                "Could not find all selected steps. Please ensure that all steps exist "
                "and that you provided to the dotpath to them. "
                f"Steps provided: {pipe_steps}."
            )
            raise e
    else:
        typer.echo("No steps selected. Executing the whole pipeline.")

    # Instantiate and execute the pipeline
    pipeline = instantiate(cfg, _convert_="all")
    pipeline.execute()


@app.command()
def mlflow_ui(
    path: str = typer.Option("mllogs/mlflow", help="Path to logs storage."),
    port: int = typer.Option(5000, help="Port on which the MLFlow UI is listening."),
    host: str = typer.Option(
        "127.0.0.1",
        help="Which host to use. Switch to '0.0.0.0' to e.g. allow for port-forwarding.",
    ),
):
    """Visualize logs with Mlflow."""
    import subprocess

    subprocess.run(f"mlflow ui --backend-store-uri {path} --port {port} --host {host}".split())


@app.command()
def mlflow_server(
    path: str = typer.Option("mllogs/mlflow", help="Path to logs storage."),
    port: int = typer.Option(5000, help="Port on which the server is listening."),
):
    """Spawn Mlflow server."""
    import subprocess

    subprocess.run(f"mlflow server --backend-store-uri {path} --port {port}".split())


@app.command()
def kill_mlflow_server(
    port: int = typer.Option(5000, help="Port on which the server is listening."),
):
    """Kill Mlflow server."""
    import subprocess

    subprocess.run(
        f"kill -9 $(lsof -t -i:{port})".split(), check=True, stderr=subprocess.DEVNULL
    )


@app.command()
def download_mlflow_data(
    tracking_uri: Annotated[
        str, typer.Option(help="The tracking URI of the MLFlow server.")
    ] = "https://mlflow.intertwin.fedcloud.eu/",
    experiment_id: Annotated[
        str, typer.Option(help="The experiment ID that you wish to retrieve data from.")
    ] = "48",
    output_file: Annotated[
        str, typer.Option(help="The file path to save the data to.")
    ] = "mlflow_data.csv",
):
    """Download metrics data from MLFlow experiments and save to a CSV file.

    Requires MLFlow authentication if the server is configured to use it.
    Authentication must be provided via the following environment variables:
    'MLFLOW_TRACKING_USERNAME' and 'MLFLOW_TRACKING_PASSWORD'.
    """

    mlflow_credentials_set = (
        "MLFLOW_TRACKING_USERNAME" in os.environ and "MLFLOW_TRACKING_PASSWORD" in os.environ
    )
    if not mlflow_credentials_set:
        typer.echo(
            "\nWarning: MLFlow authentication environment variables are not set. "
            "If the server requires authentication, your request will fail."
            "You can authenticate by setting environment variables before running:\n"
            "\texport MLFLOW_TRACKING_USERNAME=your_username\n"
            "\texport MLFLOW_TRACKING_PASSWORD=your_password\n"
        )

    import mlflow
    import pandas as pd
    from mlflow import MlflowClient

    mlflow.set_tracking_uri(tracking_uri)
    client = MlflowClient()

    # Handling authentication
    try:
        typer.echo(f"\nConnecting to MLFlow server at {tracking_uri}")
        typer.echo(f"Accessing experiment ID: {experiment_id}")
        runs = client.search_runs(experiment_ids=[experiment_id])
        typer.echo(f"Authentication successful! Found {len(runs)} runs.")
    except mlflow.MlflowException as e:
        status_code = e.get_http_status_code()
        if status_code == 401:
            typer.echo(
                "Authentication with MLFlow failed with code 401! Either your "
                "environment variables are not set or they are incorrect!"
            )
            typer.Exit()
        else:
            typer.echo(e.message)
            typer.Exit()

    all_metrics = []
    for run_idx, run in enumerate(runs):
        run_id = run.info.run_id
        metric_keys = run.data.metrics.keys()  # Get all metric names

        typer.echo(f"Processing run {run_idx + 1}/{len(runs)}")
        for metric_name in metric_keys:
            metrics = client.get_metric_history(run_id, metric_name)
            for metric in metrics:
                all_metrics.append(
                    {
                        "run_id": run_id,
                        "metric_name": metric.key,
                        "value": metric.value,
                        "step": metric.step,
                        "timestamp": metric.timestamp,
                    }
                )

    if not all_metrics:
        typer.echo("No metrics found in the runs")
        typer.Exit()

    df_metrics = pd.DataFrame(all_metrics)
    df_metrics.to_csv(output_file, index=False)
    typer.echo(f"Saved data to '{Path(output_file).resolve()}'!")


def tensorboard_ui(
    path: str = typer.Option("mllogs/tensorboard", help="Path to logs storage."),
    port: int = typer.Option(6006, help="Port on which the Tensorboard UI is listening."),
    host: str = typer.Option(
        "127.0.0.1",
        help="Which host to use. Switch to '0.0.0.0' to e.g. allow for port-forwarding.",
    ),
):
    """Visualize logs with TensorBoard."""
    import subprocess

    subprocess.run(f"tensorboard --logdir={path} --port={port} --host={host}".split())


if __name__ == "__main__":
    app()



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Anna Elisa Lappe
#
# Credit:
# - Anna Lappe <anna.elisa.lappe@cern.ch> - CERN
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# -------------------------------------------------------------------------------------

import os
import tempfile
from argparse import Namespace
from pathlib import Path
from typing import Any, Dict, Literal, Optional, Union

import yaml
from lightning.pytorch.callbacks.model_checkpoint import ModelCheckpoint
from lightning.pytorch.loggers.logger import Logger as LightningLogger
from lightning.pytorch.loggers.logger import rank_zero_experiment
from lightning.pytorch.loggers.utilities import _scan_checkpoints
from lightning.pytorch.utilities import rank_zero_only
from torch import Tensor
from typing_extensions import override

from itwinai.loggers import Logger as ItwinaiBaseLogger


class ItwinaiLogger(LightningLogger):
    """Adapter between PyTorch Lightning logger and itwinai logger.

    This adapter forwards logging calls from PyTorch Lightning to the
    itwinai Logger instance, using the itwinai Logger's `log` method.
    It supports the lightning logging of metrics, hyperparameters, and checkpoints.
    Additionally, any function calls can be forwarded to the itwinai logger instance
    though the `experiment` property of this Adapter.
    """

    def __init__(
        self,
        itwinai_logger: ItwinaiBaseLogger,
        log_model: Union[Literal["all"], bool] = False,
        skip_finalize: bool = False,
    ):
        """Initializes the adapter with an itwinai logger instance.

        Args:
            itwinai_logger (Logger): An instance of itwinai Logger.
            log_model (Union[Literal["all"], bool], optional):
                Specifies which checkpoints to log.
                If "all", logs all checkpoints; if True, logs the best k checkpoints according
                to the specifications given as `save_top_k` in the Lightning ModelCheckpoint;
                if False, does not log checkpoints.
            skip_finalize (bool): if True, do not finalize the logger in the finalize method.
                This is useful when you also want to use the logger outside of lightning.
                Defaults to False.
        """
        self.itwinai_logger = itwinai_logger
        self._log_model = log_model
        self._skip_finalize = skip_finalize
        self._logged_model_time = {}
        self._checkpoint_callback = None

    @property
    def name(self) -> Optional[str]:
        """Return the experiment name."""
        self.experiment.experiment_id

    @property
    def version(self) -> Optional[Union[int, str]]:
        """Return the experiment version."""
        self.experiment.run_id

    @property
    @override
    @rank_zero_only
    def save_dir(self) -> Optional[str]:
        """Return the directory where the logs are stored."""
        return self.experiment.savedir

    @property
    @rank_zero_experiment
    def experiment(self) -> ItwinaiBaseLogger:
        """Lightning Logger function.
        Initializes and returns the itwinai Logger context for experiment tracking.

        Returns:
            Logger: The itwinai logger instance.
        """
        if not self.itwinai_logger.is_initialized:
            # With the rank_zero_experiment decorators the rank will always be 0
            self.itwinai_logger.create_logger_context(rank=0)

        return self.itwinai_logger

    @override
    @rank_zero_only
    def finalize(self, status: str) -> None:
        """Lightning Logger function.
        Logs any remaining checkpoints and closes the logger context.

        Args:
            status (str): Describes the status of the training (e.g., 'completed', 'failed').
            The status is not needed for this function but part of the parent classes'
                (LightningLogger)
            finalize functions signature, and therefore must be propagated here.
        """
        if not self.itwinai_logger.is_initialized or self._skip_finalize:
            return

        # Log checkpoints if the last checkpoint was saved but not logged
        if self._checkpoint_callback:
            self._scan_and_log_checkpoints(self._checkpoint_callback)

        self.experiment.destroy_logger_context()

    @override
    @rank_zero_only
    def log_metrics(self, metrics: Dict[str, float], step: Optional[int] = None) -> None:
        """Lightning Logger function.
        Logs the given metrics and is usually called by the Lightning Trainer.

        Args:
            metrics (Dict[str, float]): Dictionary of metrics to log.
            step (Optional[int], optional): Training step associated with the metrics.
                Defaults to None.
        """
        for identifier, item in metrics.items():
            self.experiment.log(item=item, identifier=identifier, kind="metric", step=step)

    @override
    @rank_zero_only
    def log_hyperparams(self, params: Union[Dict[str, Any], Namespace]) -> None:
        """Lightning Logger function. Logs hyperparameters for the experiment.

        Args:
            params (Union[Dict[str, Any], Namespace]): Hyperparameters dictionary or object.
        """
        if isinstance(params, Namespace):
            params = vars(params)

        self.experiment.save_hyperparameters(params)

    @override
    def after_save_checkpoint(self, checkpoint_callback: ModelCheckpoint) -> None:
        """Lightning Logger function. Handles checkpoint saving to the logger after
        the ModelCheckpoint Callback of the Lightning Trainer is called.
        The checkpoints are logged as artifacts.

        Args:
            checkpoint_callback (ModelCheckpoint): Callback instance to manage checkpointing.
        """
        if self._log_model == "all" or (
            self._log_model is True and checkpoint_callback.save_top_k == -1
        ):
            self._scan_and_log_checkpoints(checkpoint_callback)
        elif self._log_model is True:
            self._checkpoint_callback = checkpoint_callback

    def _scan_and_log_checkpoints(self, checkpoint_callback: ModelCheckpoint) -> None:
        """Scans and logs checkpoints as artifacts in the experiment.

        This function retrieves new checkpoints, logs them as artifacts, and saves
        related metadata, including checkpoint score, filename, and any model checkpoint
        parameters.

        Args:
            checkpoint_callback (ModelCheckpoint): Callback instance used for checkpointing.
        """
        # Get checkpoints to be saved with associated score
        checkpoints = _scan_checkpoints(checkpoint_callback, self._logged_model_time)
        # Log iteratively all new checkpoints
        for time, path, score, _ in checkpoints:
            metadata = {
                "score": score.item() if isinstance(score, Tensor) else score,
                "original_filename": Path(path).name,
                checkpoint_callback.__class__.__name__: {
                    k: getattr(checkpoint_callback, k)
                    for k in [
                        "monitor",
                        "mode",
                        "save_last",
                        "save_top_k",
                        "save_weights_only",
                        "_every_n_train_steps",
                    ]
                    # ensure it does not break if `ModelCheckpoint` args change
                    if hasattr(checkpoint_callback, k)
                },
            }
            aliases = (
                ["latest", "best"]
                if path == checkpoint_callback.best_model_path
                else ["latest"]
            )

            artifact_path = Path(path).stem

            # Log the checkpoint
            self.experiment.log(item=path, identifier=artifact_path, kind="artifact")

            with tempfile.TemporaryDirectory(
                prefix="test", suffix="test", dir=os.getcwd()
            ) as tmp_dir:
                # Save the metadata
                with open(f"{tmp_dir}/metadata.yaml", "w") as tmp_file_metadata:
                    yaml.dump(metadata, tmp_file_metadata, default_flow_style=False)

                # Save the aliases
                with open(f"{tmp_dir}/aliases.txt", "w") as tmp_file_aliases:
                    tmp_file_aliases.write(str(aliases))

                # Log metadata and aliases
                self.experiment.log(item=tmp_dir, identifier=artifact_path, kind="artifact")

            self._logged_model_time[path] = time



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Jarl Sondre Sæther
#
# Credit:
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# --------------------------------------------------------------------------------------

"""This file contains the sample code that was used for the snippets in the interTwin
presentation held on Feb. 18. These code snippets are meant as outlines for how to use
itwinai to simplify distributed ML.
"""

from itwinai.torch.distributed import TorchDDPStrategy
from itwinai.torch.trainer import TorchTrainer


# Included for the sake of linting
def train(model):
    pass


##############################################################################
# Using itwinai's Strategy but not the TorchTrainer
##############################################################################

# Create and initialize strategy
strategy = TorchDDPStrategy(backend="nccl")
strategy.init()

# Create dataset as usual
train_dataset = ...

# Use 'strategy' to create dataloader
train_dataloader = strategy.create_dataloader(train_dataset, ...)

# Create model, optimizer and scheduler as usual
model, optimizer, scheduler = ...

# Distribute them using 'strategy'
model, optimizer, scheduler = strategy.distributed(model, optimizer, scheduler)

# Train model as usual
train(model)  # Note: have to notify 'strategy' every time an epoch passes

# Clean up strategy at the end
strategy.clean_up()
##############################################################################


##############################################################################
# Using itwinai's TorchTrainer (which uses Strategy internally)
##############################################################################

# Create dataset as usual
train_dataset = ...

# Create model as usual
model = ...

trainer = TorchTrainer(config={}, model=model, strategy="ddp")

_, _, _, trained_model = trainer.execute(train_dataset, ...)
##############################################################################



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import contextlib
import dataclasses
import time
from datetime import datetime
from typing import Annotated

import dagger
import yaml
from dagger import Doc, dag, function, object_type

from .k8s import create_pod_manifest


@object_type
class InterLink:
    """Wrapper for interLink service, usefult to communciate with the interLink VK."""

    values: Annotated[dagger.Secret, Doc("interLink installer configuration")]
    name: Annotated[
        str, Doc("Name of the k3s cluster in which the interLink VK is deployed")
    ] = "interlink-cluster"
    wait: Annotated[
        int,
        Doc("Sleep time (seconds) needed to wait for the VK to appear in the k3s cluster."),
    ] = 60
    vk_name: Annotated[
        str | None,
        Doc("Name of the interLink VK. Automatically detected from 'values' if not given"),
    ] = None
    kubernetes: Annotated[
        dagger.Service | None,
        Doc(
            "Endpoint to exsisting k3s service to attach to. "
            "Example (from the CLI): tcp://localhost:1997"
        ),
    ] = None

    _service: dagger.Service | None = dataclasses.field(init=False, default=None)

    @function
    async def cluster_config(
        self,
        local: Annotated[bool, Doc("Whether to access the cluster from localhost.")] = False,
    ) -> dagger.File:
        """Returns the config file for the k3s cluster."""
        k3s = dag.k3_s(self.name)

        return k3s.config(local=local)

    @function
    async def interlink_cluster(self) -> dagger.Service:
        """Get interLink VK deployed on a k3s cluster. Returns k3s server as a service.

        This is the first method you want to call from this class to setup the k3s cluster
        and deploy the VK if it does not exist yet. However, if an existing service is
        available, it will reuse it.
        """

        if not self.vk_name:
            values_dict = yaml.safe_load(await self.values.plaintext())
            self.vk_name = values_dict["nodeName"]

        if self.kubernetes:
            self._service = self.kubernetes

        if self._service:
            # Service was already initialized
            return self._service

        k3s = dag.k3_s(self.name)
        server = k3s.server()

        self._service = await server.start()

        # Deploy interLink VK in the k3s cluster
        # TODO: here I could have additional deployments in the cluster, e.g., a jupyterhub
        # deployment and they should be activated or not using simple bool args from this
        # function. Example interlink-cluster --jupyterhub
        await (
            dag.container()
            .from_("alpine/helm")
            .with_exec(["apk", "add", "kubectl"])
            .with_mounted_file("/.kube/config", k3s.config())
            .with_mounted_secret("/values.yaml", self.values)
            .with_env_variable("KUBECONFIG", "/.kube/config")
            .with_exec(
                [
                    "helm",
                    "install",
                    "--debug",
                    "my-node",
                    "oci://ghcr.io/intertwin-eu/interlink-helm-chart/interlink",
                    "--values",
                    "/values.yaml",
                ]
            )
            # Wait for the VK to appear as a node
            .with_exec(["sleep", f"{self.wait}"])
            .stdout()
        )

        await k3s.kubectl("wait --for=condition=Ready nodes --all --timeout=300s").stdout()

        return self._service

    @contextlib.asynccontextmanager
    async def start_serving(self) -> dagger.Service:
        """Start and stop interLink service."""
        yield await self.interlink_cluster()
        await self.teardown()

    @function
    async def teardown(self) -> None:
        """Stop the k3s service on which interLink VK is running.

        Returns:
            dagger.Service: k3s service.
        """
        await self._service.stop()
        self._service = None

    @function
    async def test_offloading(
        self,
        partition: Annotated[
            str, Doc("HPC partition on which to test the offloading")
        ] = "dev",
    ) -> str:
        """Test container offloading mechanism on remote HPC using interLink by running simple
        tests."""

        # Request memory and CPUs
        n_cpus = 4
        resources = {
            "limits": {"cpu": 8, "memory": "16Gi"},
            "requests": {"cpu": n_cpus, "memory": "10Gi"},
        }
        annotations = {
            "slurm-job.vk.io/flags": (
                # Use the dev partition for these simple tests
                f"-p {partition} --ntasks-per-node=1 --nodes=1 "
                f"--cpus-per-task={n_cpus} --time=00:15:00"
            )
        }
        if partition == "gpu":
            # Add requests for GPU as well, otherwise the request may hang forever
            annotations["slurm-job.vk.io/flags"] += " --gres=gpu:1 --gpus-per-node=1"
        image_path = "/ceph/hpc/data/st2301-itwin-users/cern/hello-world-image.sif"
        pod_name = "interlink-test-offloading"

        # Launch interLink service
        async with self.start_serving():
            assert self.vk_name is not None, "vk_name is none!"

            stdout = []
            stdout.append(await self.status())

            # Test #1: test successful pod
            stdout.append("### Test succesful pod ###")
            full_name = pod_name + "-test-succeeded"
            pod_manifest = create_pod_manifest(
                annotations=annotations,
                image_path=image_path,
                cmd_args=["sleep 5 && ls -la"],
                name=full_name,
                resources=resources,
                target_node=self.vk_name,
            )
            pod_manifest_str = yaml.dump(pod_manifest)

            stdout.append(await self.submit_pod(pod_manifest_str))
            # await k8s_client.container().terminal()
            status, logs = await self.wait_pod(name=full_name, timeout=300, poll_interval=5)
            assert (
                status == "Succeeded"
            ), f"Pod did not complete successfully as expected. Got status: {status}"
            stdout.extend([status, logs])

            # Test #2: test pod expected to fail
            stdout.append("### Test failing pod ###")
            full_name = pod_name + "-test-failed"
            pod_manifest = create_pod_manifest(
                annotations=annotations,
                image_path=image_path,
                cmd_args=["sleep 5 && exit 2"],
                name=full_name,
                resources=resources,
                target_node=self.vk_name,
            )
            pod_manifest_str = yaml.dump(pod_manifest)

            stdout.append(await self.submit_pod(pod_manifest_str))
            # await k8s_client.container().terminal()
            status, logs = await self.wait_pod(name=full_name, timeout=300, poll_interval=5)
            assert status == "Failed", f"Pod did not fail as expected. Got status: {status}"
            stdout.extend([status, logs])

            # Test #3: test pod expected to timeout
            stdout.append("### Test timed out pod ###")
            full_name = pod_name + "-test-timeout"
            pod_manifest = create_pod_manifest(
                annotations=annotations,
                image_path=image_path,
                cmd_args=["sleep 10"],
                name=full_name,
                resources=resources,
                target_node=self.vk_name,
            )
            pod_manifest_str = yaml.dump(pod_manifest)

            stdout.append(await self.submit_pod(pod_manifest_str))
            # await k8s_client.container().terminal()
            status, logs = await self.wait_pod(name=full_name, timeout=3, poll_interval=1)
            assert status in (
                "Pod timed out with status: Pending",
                "Pod timed out with status: Running",
            ), f"Pod was expected to run out of time but it didn't. Got status: {status}"
            stdout.extend([status, logs])

        return "\n\n".join(stdout)

    @function
    async def client(self) -> dagger.Container:
        """Returns a client for the k3s cluster. If the cluster does not exist,
        it will create it.

        >>> dagger call interlink --values=file:tmp.yaml client teminal
        """
        await self.interlink_cluster()
        return dag.k3_s(name=self.name).container()

    async def _run_cmd(self, cmd: list[str], stdin: str | None = None) -> str:
        container = await self.client()
        return await (
            container
            # Invalidate cache
            .with_env_variable("CACHE", datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f"))
            .with_exec(cmd, stdin=stdin)
            .stdout()
        )

    @function
    async def status(self) -> str:
        """Get k8s cluster status, including nodes and pods under default namespace."""
        pods = await self._run_cmd("kubectl get pods".split())
        nodes = await self._run_cmd("kubectl get nodes".split())
        return f"Nodes:\n{nodes}\n\nPods:\n{pods}"

    @function
    async def submit_pod(
        self, manifest: Annotated[str, Doc("pod manifest serialized as string.")]
    ) -> str:
        """Submit pod to k8s cluster. Returns the result of submission."""
        cmd = ["kubectl", "apply", "-f", "-"]
        return await self._run_cmd(cmd, stdin=manifest)

    @function
    async def get_pod_status(self, name: Annotated[str, Doc("pod name")]) -> str:
        """Get pod status: Pending, Running, Succeeded, Failed, Unknown.
        Returns the pod status in that moment.
        """
        cmd = [
            "kubectl",
            "get",
            "pod",
            f"{name}",
            "-n",
            "default",
            "-o",
            "jsonpath='{.status.phase}'",
        ]
        status = await self._run_cmd(cmd)
        return status.strip("'")

    @function
    async def get_pod_logs(self, name: Annotated[str, Doc("pod name")]) -> str:
        """Get pod logs."""
        cmd = [
            "kubectl",
            "logs",
            "--insecure-skip-tls-verify-backend",
            f"{name}",
            "-n",
            "default",
        ]
        return await self._run_cmd(cmd)

    @function
    async def delete_pod(self, name: Annotated[str, Doc("pod name")]) -> str:
        """Delete pod. Returns the result of ``kubectl delete pod``."""
        cmd = [
            "kubectl",
            "delete",
            "pod",
            f"{name}",
            "-n",
            "default",
            "--grace-period=0",
            "--force",
        ]
        return await self._run_cmd(cmd)

    @function
    async def wait_pod(
        self,
        name: Annotated[str, Doc("pod name")],
        timeout: Annotated[int, Doc("timeout in seconds")] = 300,
        poll_interval: Annotated[int, Doc("how often to check the pod status")] = 5,
    ) -> tuple[str, str]:
        """Wait for pod termination (Succeeded, Failed) or timeout.
        Returns last pod status and logs detected.
        """
        counter = 0
        timeout = max(int(timeout / poll_interval), 1)
        # Allow at most about 60 seconds of unk status
        unk_timeout = max(int(60 / poll_interval), 1)
        try:
            while True:
                status = await self.get_pod_status(name=name)
                logs = await self.get_pod_logs(name=name)
                if status in ["Succeeded", "Failed"]:
                    break
                counter += 1
                if counter > timeout or status == "Unknown" and counter > unk_timeout:
                    status = f"Pod timed out with status: {status}"
                    break
                time.sleep(poll_interval)
        finally:
            await self.delete_pod(name=name)
        return status, logs



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Henry Mutegeki
#
# Credit:
# - Henry Mutegeki <henry.mutegeki@cern.ch> - CERN
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import argparse

import torch
import torch.nn as nn
from torchvision import datasets, transforms

from itwinai.loggers import MLFlowLogger
from itwinai.torch.gan import GANTrainer, GANTrainingConfiguration


class Generator(nn.Module):
    def __init__(self, z_dim, g_hidden, image_channel):
        super(Generator, self).__init__()
        self.main = nn.Sequential(
            # input layer
            nn.ConvTranspose2d(z_dim, g_hidden * 8, 4, 1, 0, bias=False),
            nn.BatchNorm2d(g_hidden * 8),
            nn.ReLU(True),
            # 1st hidden layer
            nn.ConvTranspose2d(g_hidden * 8, g_hidden * 4, 4, 2, 1, bias=False),
            nn.BatchNorm2d(g_hidden * 4),
            nn.ReLU(True),
            # 2nd hidden layer
            nn.ConvTranspose2d(g_hidden * 4, g_hidden * 2, 4, 2, 1, bias=False),
            nn.BatchNorm2d(g_hidden * 2),
            nn.ReLU(True),
            # 3rd hidden layer
            nn.ConvTranspose2d(g_hidden * 2, g_hidden, 4, 2, 1, bias=False),
            nn.BatchNorm2d(g_hidden),
            nn.ReLU(True),
            # output layer
            nn.ConvTranspose2d(g_hidden, image_channel, 4, 2, 1, bias=False),
            nn.Tanh(),
        )

    def forward(self, input):
        return self.main(input)


class Discriminator(nn.Module):
    def __init__(self, d_hidden, image_channel):
        super(Discriminator, self).__init__()
        self.main = nn.Sequential(
            # 1st layer
            nn.Conv2d(image_channel, d_hidden, 4, 2, 1, bias=False),
            nn.LeakyReLU(0.2, inplace=True),
            # 2nd layer
            nn.Conv2d(d_hidden, d_hidden * 2, 4, 2, 1, bias=False),
            nn.BatchNorm2d(d_hidden * 2),
            nn.LeakyReLU(0.2, inplace=True),
            # 3rd layer
            nn.Conv2d(d_hidden * 2, d_hidden * 4, 4, 2, 1, bias=False),
            nn.BatchNorm2d(d_hidden * 4),
            nn.LeakyReLU(0.2, inplace=True),
            # 4th layer
            nn.Conv2d(d_hidden * 4, d_hidden * 8, 4, 2, 1, bias=False),
            nn.BatchNorm2d(d_hidden * 8),
            nn.LeakyReLU(0.2, inplace=True),
            # output layer
            nn.Conv2d(d_hidden * 8, 1, 4, 1, 0, bias=False),
            nn.Sigmoid(),
        )

    def forward(self, input):
        return self.main(input).view(-1, 1).squeeze(1)


def main():
    parser = argparse.ArgumentParser(description="PyTorch MNIST GAN Example")
    parser.add_argument(
        "--batch-size",
        type=int,
        default=128,
        help="input batch size for training (default: 128)",
    )
    parser.add_argument(
        "--epochs", type=int, default=15, help="number of epochs to train (default: 15)"
    )
    parser.add_argument(
        "--strategy", type=str, default="ddp", help="distributed strategy (default=ddp)"
    )
    parser.add_argument(
        "--lr", type=float, default=0.001, help="learning rate (default: 0.001)"
    )
    parser.add_argument("--seed", type=int, default=1, help="random seed (default: 1)")
    parser.add_argument(
        "--ckpt-interval",
        type=int,
        default=2,
        help="how many batches to wait before logging training status",
    )
    args = parser.parse_args()

    torch.manual_seed(args.seed)

    # Dataset
    transform = transforms.Compose(
        [
            transforms.Resize(64),
            transforms.ToTensor(),
            transforms.Normalize((0.5,), (0.5,)),
        ]
    )

    train_dataset = datasets.MNIST("../data", train=True, download=True, transform=transform)
    validation_dataset = datasets.MNIST("../data", train=False, transform=transform)

    def weights_init(m):
        classname = m.__class__.__name__
        if classname.find("Conv") != -1:
            m.weight.data.normal_(0.0, 0.02)
        elif classname.find("BatchNorm") != -1:
            m.weight.data.normal_(1.0, 0.02)
            m.bias.data.fill_(0)

    # Models
    netG = Generator(z_dim=100, g_hidden=64, image_channel=1)
    netG.apply(weights_init)
    netD = Discriminator(d_hidden=64, image_channel=1)
    netD.apply(weights_init)

    # Training configuration
    training_config = GANTrainingConfiguration(
        batch_size=args.batch_size,
        optim_generator_lr=args.lr,
        optim_discriminator_lr=args.lr,
        z_dim=100,
    )

    # Logger
    logger = MLFlowLogger(experiment_name="Distributed GAN MNIST", log_freq=10)

    # Trainer
    trainer = GANTrainer(
        config=training_config,
        epochs=args.epochs,
        discriminator=netD,
        generator=netG,
        strategy=args.strategy,
        random_seed=args.seed,
        logger=logger,
        checkpoint_every=args.ckpt_interval,
    )

    # Launch training
    _, _, _, trained_model = trainer.execute(train_dataset, validation_dataset, None)


if __name__ == "__main__":
    main()



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Anna Lappe
#
# Credit:
# - Anna Lappe <anna.elisa.lappe@cern.ch> - CERN
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

from pathlib import Path
from typing import Tuple

from torch.utils.data import Dataset, random_split
from torchvision import datasets, transforms

from itwinai.components import DataGetter, DataSplitter


class FashionMNISTGetter(DataGetter):
    def __init__(self, data_dir: str | Path = "data") -> None:
        super().__init__()
        self.data_dir = data_dir

    def execute(self) -> Dataset:
        """Load the FashionMNIST dataset from the specified directory."""
        print("Loading FashionMNIST dataset...")
        train_dataset = datasets.FashionMNIST(
            self.data_dir,
            train=True,
            download=True,
            transform=transforms.Compose(
                [transforms.ToTensor(), transforms.Normalize((0.5,), (0.5,))]
            ),
        )
        print("Loading complete!")
        return train_dataset


class FashionMNISTSplitter(DataSplitter):
    def __init__(
        self,
        train_proportion: float,
        validation_proportion: float,
        test_proportion: float,
        name: str | None = None,
    ) -> None:
        super().__init__(train_proportion, validation_proportion, test_proportion, name)

    def execute(self, dataset: Dataset) -> Tuple[Dataset, Dataset, Dataset]:
        """Split the dataset into train, validation, and test sets."""
        print("Splitting dataset...")
        total_size = len(dataset)
        train_size = int(self.train_proportion * total_size)
        validation_size = int(self.validation_proportion * total_size)
        test_size = total_size - train_size - validation_size

        train_dataset, validation_dataset, test_dataset = random_split(
            dataset, [train_size, validation_size, test_size]
        )
        print("Splitting complete!")
        return train_dataset, validation_dataset, test_dataset



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# - Rakesh Sarma <r.sarma@fz-juelich.de> - Juelich
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import os
from typing import Any, Dict, List, Literal, Optional, Union

import torch
from torch import nn
from torch.utils.data import DataLoader, Dataset

from ..components import Predictor, monitor_exec
from ..loggers import Logger
from ..serialization import ModelLoader
from .config import TrainingConfiguration
from .distributed import TorchDistributedStrategy
from .trainer import TorchTrainer
from .type import Batch


class TorchModelLoader(ModelLoader):
    """Loads a torch model from somewhere.

    Args:
        model_uri (str): Can be a path on local filesystem
            or an mlflow 'locator' in the form:
            'mlflow+MLFLOW_TRACKING_URI+RUN_ID+ARTIFACT_PATH'
    """

    def __init__(self, model_uri: str, model_class: nn.Module | None = None):
        self.model_uri = model_uri
        self.model_class = model_class

    def __call__(self) -> nn.Module:
        """Loads model from model URI.

        Raises:
            ValueError: if the model URI is not recognized
                or the model is not found.

        Returns:
            nn.Module: torch neural network.
        """
        if os.path.exists(self.model_uri):
            # Model is on local filesystem.
            checkpoint = torch.load(self.model_uri)

            if isinstance(checkpoint, dict) and "model_state_dict" in checkpoint:
                if self.model_class is None:
                    raise ValueError(
                        "model_class required to instantiate model when checkpoint is dict."
                    )
                model = self.model_class()
                model.load_state_dict(checkpoint["model_state_dict"], strict=False)
            else:
                model = checkpoint

            return model.eval()

        if self.model_uri.startswith("mlflow+"):
            # Model is on an MLFLow server
            # Form is 'mlflow+MLFLOW_TRACKING_URI+RUN_ID+ARTIFACT_PATH'
            import mlflow
            from mlflow import MlflowException

            _, tracking_uri, run_id, artifact_path = self.model_uri.split("+")
            mlflow.set_tracking_uri(tracking_uri)

            # Check that run exists
            try:
                mlflow.get_run(run_id)
            except MlflowException:
                raise ValueError(f"Run ID '{run_id}' was not found!")

            # Download model weights
            ckpt_path = mlflow.artifacts.download_artifacts(
                run_id=run_id,
                artifact_path=artifact_path,
                dst_path="tmp/",
                tracking_uri=mlflow.get_tracking_uri(),
            )
            checkpoint = torch.load(ckpt_path)

            if isinstance(checkpoint, dict) and "model_state_dict" in checkpoint:
                if self.model_class is None:
                    raise ValueError("model_class is required to instantiate the model.")
                model = self.model_class()
                model.load_state_dict(checkpoint["model_state_dict"], strict=False)
            else:
                model = checkpoint

            return model.eval()

        raise ValueError(
            "Unrecognized model URI: model may not be there! "
            f"Received model URI: {self.model_uri}"
        )


class TorchPredictor(TorchTrainer, Predictor):
    """Applies a pre-trained torch model to unseen data."""

    _strategy: TorchDistributedStrategy = None
    #: PyTorch ``DataLoader`` for inference dataset.
    inference_dataloader: DataLoader = None
    #: Pre-trained PyTorch model used to make predictions.
    model: nn.Module = None
    #: PyTorch random number generator (PRNG).
    torch_rng: torch.Generator = None
    #: itwinai ``itwinai.Logger``
    logger: Logger = None

    def __init__(
        self,
        config: Dict | TrainingConfiguration,
        model: nn.Module | ModelLoader,
        strategy: Literal["ddp", "deepspeed", "horovod"] = 'ddp',
        logger: Logger | None = None,
        checkpoints_location: str = "checkpoints",
        name: str | None = None
    ) -> None:
        if isinstance(config, dict):
            self.config = TrainingConfiguration(**config)
        else:
            self.config = config

        epochs = getattr(self.config, "epochs", 1)

        super().__init__(model=model, config=self.config, epochs=epochs, name=name)
        self.save_parameters(**self.locals2params(locals()))
        self.strategy = strategy
        self.logger = logger
        self.checkpoints_location = checkpoints_location

    def distribute_model(self) -> None:
        """Distribute the torch model with the chosen strategy."""
        if self.model is None:
            raise ValueError(
                "self.model is None! Please ensure it is set before calling this method."
            )
        distribute_kwargs = {}
        # Distributed model
        self.model, _, _ = self.strategy.distributed(
            model=self.model, optimizer=None, lr_scheduler=None, **distribute_kwargs
        )

    def create_dataloaders(self, inference_dataset: Dataset) -> DataLoader:
        """Create inference dataloader.

        Args:
            inference_dataset (Dataset): inference dataset object.

        Returns:
            DataLoader: Instance of DataLoader for the given inference dataset.
        """

        self.inference_dataloader = self.strategy.create_dataloader(
            dataset=inference_dataset,
            batch_size=self.config.batch_size,
            num_workers=self.config.num_workers_dataloader,
            pin_memory=self.config.pin_gpu_memory,
            generator=self.torch_rng,
            shuffle=self.config.shuffle_test
        )

        return self.inference_dataloader

    def predict(self) -> Dict[str, Any]:
        """Predicts or runs inference on a trained ML model.

        Returns:
            Dict[str, Any]: maps each item ID to the corresponding predicted values.
        """
        all_predictions = dict()
        for samples_ids, samples in self.inference_dataloader:
            with torch.no_grad():
                pred_batch = self.model(samples.to(self.device))
            pred_batch = self.transform_predictions(pred_batch)
            for idx, pred in zip(samples_ids, pred_batch):
                # For each item in the batch
                if pred.numel() == 1:
                    pred = pred.item()
                else:
                    pred = pred.to_dense().tolist()
                all_predictions[idx] = pred
        return all_predictions

    @monitor_exec
    def execute(
        self,
        inference_dataset: Dataset,
        model: nn.Module | None = None,
    ) -> Dict[str, Any]:
        """Applies a torch model to a dataset for inference.

        Args:
            inference_dataset (Dataset[str, Any]): each item in this dataset is a
                couple (item_unique_id, item)
            model (nn.Module, optional): torch model. Overrides the existing
                model, if given. Defaults to None.

        Returns:
            Dict[str, Any]: maps each item ID to the corresponding predicted
                value(s).
        """
        self._init_distributed_strategy()
        if model is not None:
            # Overrides existing "internal" model
            self.model = model
        elif isinstance(self.model, TorchModelLoader):
            self.model = self.model()

        self.create_dataloaders(inference_dataset=inference_dataset)

        self.distribute_model()

        if self.logger:
            self.logger.create_logger_context(rank=self.strategy.global_rank())
            hparams = self.config.model_dump()
            hparams['distributed_strategy'] = self.strategy.__class__.__name__
            self.logger.save_hyperparameters(hparams)

        all_predictions = self.predict()

        if self.logger:
            self.logger.destroy_logger_context()

        self.strategy.clean_up()

        return all_predictions

    def log(
        self,
        item: Any | List[Any],
        identifier: str | List[str],
        kind: str = 'metric',
        step: int | None = None,
        batch_idx: int | None = None,
        **kwargs
    ) -> None:
        """Log ``item`` with ``identifier`` name of ``kind`` type at ``step``
        time step.

        Args:
            item (Union[Any, List[Any]]): element to be logged (e.g., metric).
            identifier (Union[str, List[str]]): unique identifier for the
                element to log(e.g., name of a metric).
            kind (str): Type of the item to be logged. Must be one
                among the list of self.supported_types. Defaults to 'metric'.
            step (int | None): logging step. Defaults to None.
            batch_idx (int | None): DataLoader batch counter
                (i.e., batch idx), if available. Defaults to None.
        """
        if self.logger:
            self.logger.log(
                item=item,
                identifier=identifier,
                kind=kind,
                step=step,
                batch_idx=batch_idx,
                **kwargs
            )

    def transform_predictions(self, batch: Batch) -> Batch:
        """Post-process the predictions of the torch model (e.g., apply
        threshold in case of multi-label classifier).
        """


class MulticlassTorchPredictor(TorchPredictor):
    """Applies a pre-trained torch model to unseen data for multiclass classification."""

    def transform_predictions(self, batch: Batch) -> Batch:
        batch = batch.argmax(-1)
        return batch


class MultilabelTorchPredictor(TorchPredictor):
    """Applies a pre-trained torch model to unseen data for
    multilabel classification, applying a threshold on the
    output of the neural network.
    """

    #: Threshold to transform probabilities into class predictions.
    #: Defaults to 0.5.
    threshold: float = 0.5

    def __init__(
        self,
        model: Union[nn.Module, ModelLoader],
        test_dataloader_class: str = "torch.utils.data.DataLoader",
        test_dataloader_kwargs: Optional[Dict] = None,
        threshold: float = 0.5,
        name: str = None,
    ) -> None:
        super().__init__(model, test_dataloader_class, test_dataloader_kwargs, name)
        self.threshold = threshold

    def transform_predictions(self, batch: Batch) -> Batch:
        return (batch > self.threshold).float()


class RegressionTorchPredictor(TorchPredictor):
    """Applies a pre-trained torch model to unseen data for
    regression, leaving untouched the output of the neural
    network.
    """

    def transform_predictions(self, batch: Batch) -> Batch:
        return batch



import inspect

from itwinai.cli import generate_slurm
from itwinai.slurm.utils import get_slurm_job_parser


def test_cli_slurm_function_signature():
    """Test that function signature in cli.py matches argparser"""
    args = inspect.getfullargspec(generate_slurm).args
    parser = get_slurm_job_parser()

    ignored_args = ["print_config", "help"]
    parser_args = {arg.dest for arg in parser._actions}
    parser_args -= set(ignored_args)

    missing_in_function = parser_args - set(args)
    missing_in_parser = set(args) - parser_args

    assert not missing_in_function and not missing_in_parser, (
        f"Arguments missing in function: {missing_in_function}, "
        f"Arguments missing in parser: {missing_in_parser}"
    )



import re
from typing import Dict, List


def add_lowest_library_function(
    stack_traces: List[List[Dict]], library_name: str = "itwinai"
) -> None:
    """Iterates through each stack traces in the given list and finds the lowest call with
    the given library name, if any, and adds it to each dictionary in the trace.

    This function changes the given list in-place.
    """
    trace: List[Dict]
    for trace in stack_traces:
        library_path = "Not Found"
        library_function = "Not Found"
        library_line = "Not Found"
        for function_dict in reversed(trace):
            if library_name not in function_dict["path"]:
                continue

            library_path = function_dict["path"]
            library_function = function_dict["name"]
            library_line = function_dict["line"]
            # Since we iterate backwards then we stop once we found a match
            break

        # Adding the last one that was found (lowest) to all the dicts
        for function_dict in trace:
            function_dict["library_function_name"] = library_function
            function_dict["library_function_path"] = library_path
            function_dict["library_function_line"] = library_line


def get_aggregated_paths(functions: List[Dict]) -> List[Dict]:
    """Return a new list where all entries with the same information have been combined
    into one. This means that if the name, path and line for the function itself and the
    specified library function are the same in two entries, then they are combined into one,
    with their number of samples summed.
    """
    aggregated_functions = {}
    for entry in functions:
        key = (
            entry["name"],
            entry["path"],
            entry["line"],
            entry["library_function_name"],
            entry["library_function_path"],
            entry["library_function_line"],
        )
        aggregated_functions[key] = aggregated_functions.get(key, 0) + entry["num_samples"]
    aggregated_functions = [
        {
            "name": key[0],
            "path": key[1],
            "line": key[2],
            "library_function_name": key[3],
            "library_function_path": key[4],
            "library_function_line": key[5],
            "num_samples": value,
        }
        for key, value in aggregated_functions.items()
    ]
    return aggregated_functions


def parse_trace_line_to_dict(trace_line: str, num_samples: int) -> Dict[str, str | int]:
    """Parses a single trace line, which contains a function name, file name and line number,
    to a dictionary with the name, path, line and number of samples.

    Raises:
        ValueError: If the given trace line does not conform to the expected structure of
        "function_name (path/to/function:line_number)"
    """
    function_pattern = r"([^\s]+) \(([^():]+):(\d+)\)"
    pattern_match = re.match(function_pattern, trace_line)
    if pattern_match is None:
        raise ValueError(
            f"Failed to extract function information from trace line: '{trace_line}'."
        )
    function_name, function_path, line = pattern_match.groups()

    if "frozen importlib" in function_path:
        function_path = "built-in"
        line = "-1"

    return {
        "name": function_name,
        "path": function_path,
        "line": line,
        "num_samples": num_samples,
    }


def convert_stack_trace_to_list(line: str) -> List[Dict[str, str | int]]:
    """Converts a single line of the raw profiling output, which contains a stack trace
    and a number of samples, to a list of dictionaries with the data for each line of the
    stack trace.

    Note:
        The number of samples is the last number in the string, separated from the rest with
        a space. There can be multiple spaces, however, so we have to extract the last element.

    Raises:
        ValueError: If any of the samples are not numeric.
        ValueError: If any of the lines in the stack trace are not formatted as expected.
    """
    num_samples_str = line.split(" ")[-1].strip()

    if not num_samples_str.isnumeric():
        raise ValueError(
            f'Extracted number of samples, "{num_samples_str}", is not all numeric. Was part'
            f'of line: "{line}".'
        )
    num_samples = int(num_samples_str)

    # Everything except the 'num_samples'
    stack_trace = " ".join(line.split(" ")[:-1]).strip()
    if not stack_trace:
        return []

    # The various function calls are separated with semicolons
    stack_trace = stack_trace.split(";")
    stack_trace = [trace_line for trace_line in stack_trace if trace_line]

    result = []
    for trace_line in stack_trace:
        try:
            trace_line_dict = parse_trace_line_to_dict(trace_line, num_samples)
            result.append(trace_line_dict)
        except ValueError as exception:
            raise ValueError(
                f"Failed to convert line from profiling output. Line was: \n\n{line}\nError "
                f"was: {str(exception)}"
            )

    return result
